# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Test Data Plugin for DataLab
----------------------------

This plugin is an example of DataLab plugin. It provides test data samples
and some actions to test DataLab functionalities.
"""

import cdl.obj as dlo
import cdl.tests.data as test_data
from cdl.computation import image as cpima
from cdl.computation import signal as cpsig
from cdl.config import _
from cdl.plugins import PluginBase, PluginInfo

# ------------------------------------------------------------------------------
# All computation functions must be defined as global functions, otherwise
# they cannot be pickled and sent to the worker process
# ------------------------------------------------------------------------------


def add_noise_to_signal(
    src: dlo.SignalObj, p: test_data.GaussianNoiseParam
) -> dlo.SignalObj:
    """Add gaussian noise to signal"""
    dst = cpsig.dst_11(src, "add_gaussian_noise", f"mu={p.mu},sigma={p.sigma}")
    test_data.add_gaussian_noise_to_signal(dst, p)
    return dst


def add_noise_to_image(src: dlo.ImageObj, p: dlo.NormalRandomParam) -> dlo.ImageObj:
    """Add gaussian noise to image"""
    dst = cpima.dst_11(src, "add_gaussian_noise", f"mu={p.mu},sigma={p.sigma}")
    test_data.add_gaussian_noise_to_image(dst, p)
    return dst


class PluginTestData(PluginBase):
    """DataLab Test Data Plugin"""

    PLUGIN_INFO = PluginInfo(
        name=_("Test data"),
        version="1.0.0",
        description=_("Testing DataLab functionalities"),
    )

    # Signal processing features ------------------------------------------------
    def add_noise_to_signal(self) -> None:
        """Add noise to signal"""
        self.signalpanel.processor.compute_11(
            add_noise_to_signal,
            paramclass=test_data.GaussianNoiseParam,
            title=_("Add noise"),
        )

    def create_paracetamol_signal(self) -> None:
        """Create paracetamol signal"""
        obj = test_data.create_paracetamol_signal()
        self.proxy.add_object(obj)

    def create_noisy_signal(self) -> None:
        """Create noisy signal"""
        obj = self.signalpanel.new_object(add_to_panel=False)
        if obj is not None:
            noiseparam = test_data.GaussianNoiseParam(_("Noise"))
            self.signalpanel.processor.update_param_defaults(noiseparam)
            if noiseparam.edit(self.main):
                test_data.add_gaussian_noise_to_signal(obj, noiseparam)
                self.proxy.add_object(obj)

    # Image processing features ------------------------------------------------
    def add_noise_to_image(self) -> None:
        """Add noise to image"""
        self.imagepanel.processor.compute_11(
            add_noise_to_image,
            paramclass=dlo.NormalRandomParam,
            title=_("Add noise"),
        )

    def create_peak2d_image(self) -> None:
        """Create 2D peak image"""
        obj = self.imagepanel.new_object(add_to_panel=False)
        if obj is not None:
            param = test_data.PeakDataParam.create(size=max(obj.data.shape))
            self.imagepanel.processor.update_param_defaults(param)
            if param.edit(self.main):
                obj.data = test_data.get_peak2d_data(param)
                self.proxy.add_object(obj)

    def create_sincos_image(self) -> None:
        """Create 2D sin cos image"""
        newparam = self.edit_new_image_parameters(hide_image_type=True)
        if newparam is not None:
            obj = test_data.create_sincos_image(newparam)
            self.proxy.add_object(obj)

    def create_noisygauss_image(self) -> None:
        """Create 2D noisy gauss image"""
        newparam = self.edit_new_image_parameters(hide_image_type=True)
        if newparam is not None:
            obj = test_data.create_noisygauss_image(newparam, add_annotations=False)
            self.proxy.add_object(obj)

    def create_multigauss_image(self) -> None:
        """Create 2D multi gauss image"""
        newparam = self.edit_new_image_parameters(hide_image_type=True)
        if newparam is not None:
            obj = test_data.create_multigauss_image(newparam)
            self.proxy.add_object(obj)

    def create_2dstep_image(self) -> None:
        """Create 2D step image"""
        newparam = self.edit_new_image_parameters(hide_image_type=True)
        if newparam is not None:
            obj = test_data.create_2dstep_image(newparam)
            self.proxy.add_object(obj)

    def create_ring_image(self) -> None:
        """Create 2D ring image"""
        param = test_data.RingParam(_("Ring"))
        if param.edit(self.main):
            obj = test_data.create_ring_image(param)
            self.proxy.add_object(obj)

    def create_annotated_image(self) -> None:
        """Create annotated image"""
        obj = test_data.create_annotated_image()
        self.proxy.add_object(obj)

    # Plugin menu entries ------------------------------------------------------
    def create_actions(self) -> None:
        """Create actions"""
        # Signal Panel ----------------------------------------------------------
        sah = self.signalpanel.acthandler
        with sah.new_menu(_("Test data")):
            sah.new_action(_("Add noise to signal"), triggered=self.add_noise_to_signal)
            sah.new_action(
                _("Load spectrum of paracetamol"),
                triggered=self.create_paracetamol_signal,
                select_condition="always",
                separator=True,
            )
            sah.new_action(
                _("Create noisy signal"),
                triggered=self.create_noisy_signal,
                select_condition="always",
            )
        # Image Panel -----------------------------------------------------------
        iah = self.imagepanel.acthandler
        with iah.new_menu(_("Test data")):
            iah.new_action(_("Add noise to image"), triggered=self.add_noise_to_image)
            # with iah.new_menu(_("Data samples")):
            iah.new_action(
                _("Create image with peaks"),
                triggered=self.create_peak2d_image,
                select_condition="always",
                separator=True,
            )
            iah.new_action(
                _("Create 2D sin cos image"),
                triggered=self.create_sincos_image,
                select_condition="always",
            )
            iah.new_action(
                _("Create 2D noisy gauss image"),
                triggered=self.create_noisygauss_image,
                select_condition="always",
            )
            iah.new_action(
                _("Create 2D multi gauss image"),
                triggered=self.create_multigauss_image,
                select_condition="always",
            )
            iah.new_action(
                _("Create annotated image"),
                triggered=self.create_annotated_image,
                select_condition="always",
            )
            iah.new_action(
                _("Create 2D step image"),
                triggered=self.create_2dstep_image,
                select_condition="always",
            )
            iah.new_action(
                _("Create ring image"),
                triggered=self.create_ring_image,
                select_condition="always",
            )



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab I/O signal formats
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import scipy.io as sio

from cdl.config import _
from cdl.core.io.base import FormatInfo
from cdl.core.io.conv import convert_array_to_standard_type
from cdl.core.io.signal import funcs
from cdl.core.io.signal.base import SignalFormatBase

if TYPE_CHECKING:
    from cdl.core.model.signal import SignalObj
    from cdl.utils.qthelpers import CallbackWorker


class CSVSignalFormat(SignalFormatBase):
    """Object representing a CSV signal file type"""

    FORMAT_INFO = FormatInfo(
        name=_("CSV files"),
        extensions="*.csv *.txt",
        readable=True,
        writeable=True,
    )

    def read(
        self, filename: str, worker: CallbackWorker | None = None
    ) -> list[SignalObj]:
        """Read list of signal objects from file

        Args:
            filename: File name
            worker: Callback worker object

        Returns:
            List of signal objects
        """
        xydata, xlabel, xunit, ylabels, yunits, header = funcs.read_csv(
            filename, worker
        )
        if ylabels:
            # If y labels are present, we are sure that the data contains at least
            # two columns (x and y)
            objs = []
            for i, (ylabel, yunit) in enumerate(zip(ylabels, yunits)):
                obj = self.create_object(filename, i if len(ylabels) > 1 else None)
                obj.set_xydata(xydata[:, 0], xydata[:, i + 1])
                obj.xlabel = xlabel or ""
                obj.xunit = xunit or ""
                obj.ylabel = ylabel or ""
                obj.yunit = yunit or ""
                if header:
                    obj.metadata[self.HEADER_KEY] = header
                objs.append(obj)
            return objs
        return self.create_signals_from(xydata, filename)

    def write(self, filename: str, obj: SignalObj) -> None:
        """Write data to file

        Args:
            filename: Name of file to write
            obj: Signal object to read data from
        """
        funcs.write_csv(
            filename,
            obj.xydata,
            obj.xlabel,
            obj.xunit,
            [obj.ylabel],
            [obj.yunit],
            obj.metadata.get(self.HEADER_KEY, ""),
        )


class NumPySignalFormat(SignalFormatBase):
    """Object representing a NumPy signal file type"""

    FORMAT_INFO = FormatInfo(
        name=_("NumPy binary files"),
        extensions="*.npy",
        readable=True,
        writeable=True,
    )  # pylint: disable=duplicate-code

    def read_xydata(self, filename: str) -> np.ndarray:
        """Read data and metadata from file, write metadata to object, return xydata

        Args:
            filename: Name of file to read

        Returns:
            NumPy array xydata
        """
        return convert_array_to_standard_type(np.load(filename))

    def write(self, filename: str, obj: SignalObj) -> None:
        """Write data to file

        Args:
            filename: Name of file to write
            obj: Signal object to read data from
        """
        np.save(filename, obj.xydata.T)


class MatSignalFormat(SignalFormatBase):
    """Object representing a MAT-File .mat signal file type"""

    FORMAT_INFO = FormatInfo(
        name=_("MAT-Files"),
        extensions="*.mat",
        readable=True,
        writeable=True,
    )  # pylint: disable=duplicate-code

    def read(
        self, filename: str, worker: CallbackWorker | None = None
    ) -> list[SignalObj]:
        """Read data and metadata from file, write metadata to object, return xydata

        Args:
            filename: Name of file to read

        Returns:
            NumPy array xydata
        """
        mat = sio.loadmat(filename)
        allsig: list[SignalObj] = []
        for dname, data in mat.items():
            if dname.startswith("__") or not isinstance(data, np.ndarray):
                continue
            for sig in self.create_signals_from(data.squeeze(), filename):
                if dname != "sig":
                    sig.title += f" ({dname})"
                allsig.append(sig)
        return allsig

    def write(self, filename: str, obj: SignalObj) -> None:
        """Write data to file

        Args:
            filename: Name of file to write
            obj: Signal object to read data from
        """
        # metadata cannot be saved as such as their type will be lost and
        # cause problems when reading the file back
        sio.savemat(filename, {"sig": obj.xydata.T})



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab image I/O features
"""

# pylint: disable=unused-import
import cdl.core.io.image.formats  # noqa: F401
from cdl.core.io.image.base import ImageIORegistry  # noqa: F401



# -*- coding: utf-8 -*-
"""
Tutorial 'Working with Spyder'
==============================

This tutorial shows how to use the DataLab remotely from a Python script running
outside DataLab (e.g. in Spyder).
"""

# ---- Begin of the example code ----
# %% Connecting to DataLab current session

from cdlclient import SimpleRemoteProxy

proxy = SimpleRemoteProxy()
proxy.connect()

# %% Visualizing 1D data from my work

from my_work import test_my_1d_algorithm

x, y = test_my_1d_algorithm()  # Here is all my research/technical work!
proxy.add_signal("My 1D result data", x, y)  # Let's visualize it in DataLab
proxy.compute_wiener()  # Denoise the signal using the Wiener filter


# %% Visualizing 2D data from my work

from my_work import test_my_2d_algorithm

z = test_my_2d_algorithm()[1]  # Here is all my research/technical work!
proxy.add_image("My 2D result data", z)  # Let's visualize it in DataLab



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Image offset correction unit test.
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code
# guitest: show

from __future__ import annotations

import numpy as np
import pytest
from guidata.qthelpers import exec_dialog, qt_app_context

import cdl.computation.image as cpi
import cdl.param
from cdl.env import execenv
from cdl.obj import ROI2DParam
from cdl.tests.data import create_noisygauss_image
from cdl.utils.vistools import view_images_side_by_side
from cdl.widgets.imagebackground import ImageBackgroundDialog


def test_image_offset_correction_interactive() -> None:
    """Image offset correction interactive test."""
    with qt_app_context():
        i1 = create_noisygauss_image()
        dlg = ImageBackgroundDialog(i1)
        with execenv.context(delay=200):
            # On Windows, the `QApplication.processEvents()` introduced with
            # guidata V3.5.1 in `exec_dialog` is sufficient to force an update
            # of the dialog. The delay is not required.
            # On Linux, the delay is required to ensure that the dialog is displayed
            # because the `QApplication.processEvents()` do not trigger the drawing
            # event on the dialog as expected. So, the `RangeComputation2d` is not
            # drawn, the background value is not computed, and `get_rect_coords()`
            # returns `None` which causes the test to fail.
            ok = exec_dialog(dlg)
        if ok:
            param = ROI2DParam()
            ix0, iy0, ix1, iy1 = i1.physical_to_indices(dlg.get_rect_coords())
            param.x0, param.y0, param.dx, param.dy = ix0, iy0, ix1 - ix0, iy1 - iy0
            i2 = cpi.compute_offset_correction(i1, param)
            i3 = cpi.compute_clip(i2, cdl.param.ClipParam.create(lower=0))
            view_images_side_by_side(
                [i1.make_item(), i3.make_item()],
                titles=["Original image", "Corrected image"],
                title="Image offset correction and thresholding",
            )


@pytest.mark.validation
def test_image_offset_correction() -> None:
    """Image offset correction validation test."""
    i1 = create_noisygauss_image()
    param = ROI2DParam.create(x0=0, y0=0, dx=10, dy=10)
    i2 = cpi.compute_offset_correction(i1, param)

    # Check that the offset correction has been applied
    x0, y0 = param.x0, param.y0
    x1, y1 = x0 + param.dx, y0 + param.dy
    offset = np.mean(i1.data[y0:y1, x0:x1])
    assert np.allclose(i2.data, i1.data - offset), "Offset correction failed"


if __name__ == "__main__":
    test_image_offset_correction_interactive()
    test_image_offset_correction()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Proxy objects (:mod:`cdl.proxy`)
--------------------------------

The :mod:`cdl.proxy` module provides a way to access DataLab features from a proxy
class.

The list of compute methods accessible from the proxy objects is available in the
:ref:`processor_methods` section.

Remote proxy
^^^^^^^^^^^^

The remote proxy is used when DataLab is started from a different process than the
proxy. In this case, the proxy connects to DataLab XML-RPC server.

.. autoclass:: RemoteProxy
    :members:
    :inherited-members:

Local proxy
^^^^^^^^^^^

The local proxy is used when DataLab is started from the same process as the proxy.
In this case, the proxy is directly connected to DataLab main window instance. The
typical use case is high-level scripting.

.. autoclass:: LocalProxy
    :members:
    :inherited-members:

Proxy context manager
^^^^^^^^^^^^^^^^^^^^^

The proxy context manager is a convenient way to handle proxy creation and
destruction. It is used as follows:

.. code-block:: python

        with proxy_context("local") as proxy:
            proxy.add_signal(...)

The proxy type can be "local" or "remote". For remote proxy, the port can be
specified as "remote:port".

.. note:: The proxy context manager allows to use the proxy in various contexts
    (Python script, Jupyter notebook, etc.). It also allows to switch seamlessly
    between local and remote proxy, keeping the same code inside the context.

.. autofunction:: proxy_context

.. _processor_methods:

Calling processor methods using proxy objects
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

All the proxy objects provide access to the DataLab computing methods exposed by
the processor classes:

- :class:`cdl.core.gui.processor.signal.SignalProcessor`
- :class:`cdl.core.gui.processor.image.ImageProcessor`

.. seealso::

    The list of processor methods is available in tables below.

There are two ways to call a processor method:

1. Using the :meth:`calc` method of the proxy object:

.. code-block:: python

    # Call a method without parameter
    proxy.calc("compute_average")

    # This is equivalent to:
    proxy.calc("average")

    # Call a method with parameters
    p = cdl.param.MovingAverageParam.create(n=30)
    proxy.calc("compute_moving_average", p)

2. Directly calling the processor method from the proxy object:

.. code-block:: python

    # Call a method without parameter
    proxy.compute_average()

    # Call a method with parameters
    p = cdl.param.MovingAverageParam.create(n=30)
    proxy.compute_moving_average(p)

.. warning::

    The `compute_{name}` methods are not statically defined in the proxy classes (and
    not even dynamically). They are nevertheless available through the proxy objects
    thanks to the magic method :meth:`__getattr__` which forwards the call to the
    :meth:`calc` method. However, this means that the methods are not listed in the
    proxy classes documentation, and they are not available in the auto-completion
    feature of your IDE.

Number of compute methods
*************************

.. csv-table:: Number of compute methods
   :file: ../doc/processor_methods_nb.csv
   :header: Signal, Image, Total

Signal processing
*****************

The following table lists the signal processor methods - it is automatically
generated from the source code:

.. csv-table:: Signal processor methods
   :file: ../doc/processor_methods_signal.csv
   :header: Compute method, Description
   :widths: 40, 60

Image processing
****************

The following table lists the image processor methods - it is automatically
generated from the source code:

.. csv-table:: Image processor methods
    :file: ../doc/processor_methods_image.csv
    :header: Compute method, Description
    :widths: 40, 60
"""

from __future__ import annotations

from collections.abc import Generator
from contextlib import contextmanager

import guidata.dataset as gds
import numpy as np

from cdl.core.baseproxy import BaseProxy
from cdl.core.remote import RemoteClient
from cdl.obj import ImageObj, SignalObj
from cdl.utils import qthelpers as qth


class RemoteProxy(RemoteClient):
    """DataLab remote proxy class.

    This class provides access to DataLab features from a proxy class. This is the
    remote version of proxy, which is used when DataLab is started from a different
    process than the proxy.

    Args:
        autoconnect (bool): Automatically connect to DataLab XML-RPC server.

    Raises:
        ConnectionRefusedError: Unable to connect to DataLab
        ValueError: Invalid timeout (must be >= 0.0)
        ValueError: Invalid number of retries (must be >= 1)

    .. note::

        The proxy object also allows to access DataLab computing methods exposed by
        the processor classes (see :ref:`processor_methods`).

    Examples:
        Here is a simple example of how to use RemoteProxy in a Python script
        or in a Jupyter notebook:

        >>> from cdl.proxy import RemoteProxy
        >>> proxy = RemoteProxy()
        Connecting to DataLab XML-RPC server...OK (port: 28867)
        >>> proxy.get_version()
        '1.0.0'
        >>> proxy.add_signal("toto", np.array([1., 2., 3.]), np.array([4., 5., -1.]))
        True
        >>> proxy.get_object_titles()
        ['toto']
        >>> proxy["toto"]  # from title
        <cdl.core.model.signal.SignalObj at 0x7f7f1c0b4a90>
        >>> proxy[1]  # from number
        <cdl.core.model.signal.SignalObj at 0x7f7f1c0b4a90>
        >>> proxy[1].data
        array([1., 2., 3.])
        >>> proxy.set_current_panel("image")
    """

    def __init__(self, autoconnect: bool = True) -> None:
        super().__init__()
        if autoconnect:
            self.connect()


class LocalProxy(BaseProxy):
    """DataLab local proxy class.

    This class provides access to DataLab features from a proxy class. This is the
    local version of proxy, which is used when DataLab is started from the same
    process as the proxy.

    Args:
        cdl (CDLMainWindow): CDLMainWindow instance.

    .. note::

        The proxy object also allows to access DataLab computing methods exposed by
        the processor classes (see :ref:`processor_methods`).
    """

    def add_signal(
        self,
        title: str,
        xdata: np.ndarray,
        ydata: np.ndarray,
        xunit: str | None = None,
        yunit: str | None = None,
        xlabel: str | None = None,
        ylabel: str | None = None,
    ) -> bool:  # pylint: disable=too-many-arguments
        """Add signal data to DataLab.

        Args:
            title (str): Signal title
            xdata (numpy.ndarray): X data
            ydata (numpy.ndarray): Y data
            xunit (str | None): X unit. Defaults to None.
            yunit (str | None): Y unit. Defaults to None.
            xlabel (str | None): X label. Defaults to None.
            ylabel (str | None): Y label. Defaults to None.

        Returns:
            bool: True if signal was added successfully, False otherwise

        Raises:
            ValueError: Invalid xdata dtype
            ValueError: Invalid ydata dtype
        """
        return self._cdl.add_signal(title, xdata, ydata, xunit, yunit, xlabel, ylabel)

    def add_image(
        self,
        title: str,
        data: np.ndarray,
        xunit: str | None = None,
        yunit: str | None = None,
        zunit: str | None = None,
        xlabel: str | None = None,
        ylabel: str | None = None,
        zlabel: str | None = None,
    ) -> bool:  # pylint: disable=too-many-arguments
        """Add image data to DataLab.

        Args:
            title (str): Image title
            data (numpy.ndarray): Image data
            xunit (str | None): X unit. Defaults to None.
            yunit (str | None): Y unit. Defaults to None.
            zunit (str | None): Z unit. Defaults to None.
            xlabel (str | None): X label. Defaults to None.
            ylabel (str | None): Y label. Defaults to None.
            zlabel (str | None): Z label. Defaults to None.

        Returns:
            bool: True if image was added successfully, False otherwise

        Raises:
            ValueError: Invalid data dtype
        """
        return self._cdl.add_image(
            title, data, xunit, yunit, zunit, xlabel, ylabel, zlabel
        )

    def add_object(self, obj: SignalObj | ImageObj) -> None:
        """Add object to DataLab.

        Args:
            obj (SignalObj | ImageObj): Signal or image object
        """
        self._cdl.add_object(obj)

    def calc(self, name: str, param: gds.DataSet | None = None) -> None:
        """Call compute function ``name`` in current panel's processor.

        Args:
            name: Compute function name
            param: Compute function parameter. Defaults to None.

        Raises:
            ValueError: unknown function
        """
        return self._cdl.calc(name, param)

    def get_object(
        self,
        nb_id_title: int | str | None = None,
        panel: str | None = None,
    ) -> SignalObj | ImageObj:
        """Get object (signal/image) from index.

        Args:
            nb_id_title: Object number, or object id, or object title.
             Defaults to None (current object).
            panel: Panel name. Defaults to None (current panel).

        Returns:
            Object

        Raises:
            KeyError: if object not found
        """
        return self._cdl.get_object(nb_id_title, panel)

    def get_object_shapes(
        self,
        nb_id_title: int | str | None = None,
        panel: str | None = None,
    ) -> list:
        """Get plot item shapes associated to object (signal/image).

        Args:
            nb_id_title: Object number, or object id, or object title.
             Defaults to None (current object).
            panel: Panel name. Defaults to None (current panel).

        Returns:
            List of plot item shapes
        """
        return self._cdl.get_object_shapes(nb_id_title, panel)

    def add_annotations_from_items(
        self, items: list, refresh_plot: bool = True, panel: str | None = None
    ) -> None:
        """Add object annotations (annotation plot items).

        Args:
            items (list): annotation plot items
            refresh_plot (bool | None): refresh plot. Defaults to True.
            panel (str | None): panel name (valid values: "signal", "image").
                If None, current panel is used.
        """
        self._cdl.add_annotations_from_items(items, refresh_plot, panel)


@contextmanager
def proxy_context(what: str) -> Generator[LocalProxy | RemoteProxy, None, None]:
    """Context manager handling CDL proxy creation and destruction.

    Args:
        what (str): proxy type ("local" or "remote")
         For remote proxy, the port can be specified as "remote:port"

    Yields:
        Generator[LocalProxy | RemoteProxy, None, None]: proxy
            LocalProxy if what == "local"
            RemoteProxy if what == "remote" or "remote:port"

    Example:
        with proxy_context("local") as proxy:
            proxy.add_signal(...)
    """
    assert what == "local" or what.startswith("remote"), "Invalid proxy type"
    port = None
    if ":" in what:
        port = int(what.split(":")[1].strip())
    if what == "local":
        # pylint: disable=import-outside-toplevel, cyclic-import
        from cdl.core.gui.main import CDLMainWindow

        with qth.cdl_app_context(exec_loop=True):
            try:
                win = CDLMainWindow()
                proxy = LocalProxy(win)
                win.show()
                yield proxy
            finally:
                pass
    else:
        try:
            proxy = RemoteProxy(autoconnect=False)
            proxy.connect(port)
            yield proxy
        finally:
            proxy.disconnect()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab HDF5 browser module
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

import abc
import os
import os.path as osp
from typing import TYPE_CHECKING, Any, Callable

from guidata.qthelpers import (
    add_actions,
    create_action,
    create_toolbutton,
    exec_dialog,
    get_icon,
    get_std_icon,
    win32_fix_title_bar_background,
)
from guidata.widgets.arrayeditor import ArrayEditor
from plotpy.builder import make
from plotpy.plot import PlotOptions, PlotWidget
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qtpy import QtWidgets as QW
from qtpy.compat import getopenfilename

from cdl.config import _
from cdl.core.io.h5 import H5Importer
from cdl.core.model.signal import CURVESTYLES
from cdl.obj import ImageObj, SignalObj
from cdl.utils.qthelpers import qt_handle_error_message
from cdl.utils.strings import to_string

if TYPE_CHECKING:
    from plotpy.plot import BasePlot

    from cdl.core.io.h5.common import BaseNode


class AbstractTreeWidgetMeta(type(QW.QTreeWidget), abc.ABCMeta):
    """Mixed metaclass to avoid conflicts"""


class AbstractTreeWidget(QW.QTreeWidget, metaclass=AbstractTreeWidgetMeta):
    """One-column tree widget with context menu, ..."""

    def __init__(self, parent: QW.QWidget) -> None:
        super().__init__(parent)
        self.setItemsExpandable(True)
        self.itemActivated.connect(self.activated)
        self.itemClicked.connect(self.clicked)
        # Setup context menu
        self.menu = QW.QMenu(self)
        self.collapse_all_action = None
        self.collapse_selection_action = None
        self.expand_all_action = None
        self.expand_selection_action = None
        self.common_actions = self.setup_common_actions()

        self.itemSelectionChanged.connect(self.item_selection_changed)
        self.item_selection_changed()

    @abc.abstractmethod
    def activated(self, item: QW.QTreeWidgetItem) -> None:
        """Double-click event"""

    @abc.abstractmethod
    def clicked(self, item: QW.QTreeWidgetItem) -> None:
        """Item was clicked"""

    @abc.abstractmethod
    def get_actions_from_items(
        self, items: list[QW.QTreeWidgetItem]
    ) -> list[QW.QAction]:
        """Get actions from item"""
        # Right here: add other actions if necessary (reimplement this method)
        return []

    def setup_common_actions(self) -> list[QW.QAction]:
        """Setup context menu common actions"""
        self.collapse_all_action = create_action(
            self,
            _("Collapse all"),
            icon=get_icon("collapse.svg"),
            triggered=self.collapseAll,
        )
        self.expand_all_action = create_action(
            self, _("Expand all"), icon=get_icon("expand.svg"), triggered=self.expandAll
        )
        self.restore_action = create_action(
            self,
            _("Restore"),
            tip=_("Restore original tree layout"),
            icon=get_icon("restore.svg"),
            triggered=self.restore,
        )
        self.collapse_selection_action = create_action(
            self,
            _("Collapse selection"),
            icon=get_icon("collapse_selection.svg"),
            triggered=self.collapse_selection,
        )
        self.expand_selection_action = create_action(
            self,
            _("Expand selection"),
            icon=get_icon("expand_selection.svg"),
            triggered=self.expand_selection,
        )
        return [
            self.collapse_all_action,
            self.expand_all_action,
            self.restore_action,
            None,
            self.collapse_selection_action,
            self.expand_selection_action,
        ]

    def update_menu(self) -> None:
        """Update context menu"""
        self.menu.clear()
        items = self.selectedItems()
        actions = self.get_actions_from_items(items)
        if actions:
            actions.append(None)
        actions += self.common_actions
        add_actions(self.menu, actions)

    def restore(self) -> None:
        """Restore tree state"""
        self.collapseAll()
        for item in self.get_top_level_items():
            self.expandItem(item)

    def __expand_item(self, item: QW.QTreeWidgetItem) -> None:  # pragma: no cover
        """Expand item tree branch"""
        self.expandItem(item)
        for index in range(item.childCount()):
            child = item.child(index)
            self.__expand_item(child)

    def expand_selection(self) -> None:  # pragma: no cover
        """Expand selection"""
        items = self.selectedItems()
        if not items:
            items = self.get_top_level_items()
        for item in items:
            self.__expand_item(item)
        if items:
            self.scrollToItem(items[0])

    def __collapse_item(self, item: QW.QTreeWidgetItem) -> None:  # pragma: no cover
        """Collapse item tree branch"""
        self.collapseItem(item)
        for index in range(item.childCount()):
            child = item.child(index)
            self.__collapse_item(child)

    def collapse_selection(self) -> None:  # pragma: no cover
        """Collapse selection"""
        items = self.selectedItems()
        if not items:
            items = self.get_top_level_items()
        for item in items:
            self.__collapse_item(item)
        if items:
            self.scrollToItem(items[0])

    def item_selection_changed(self) -> None:
        """Item selection has changed"""
        is_selection = len(self.selectedItems()) > 0
        self.expand_selection_action.setEnabled(is_selection)
        self.collapse_selection_action.setEnabled(is_selection)

    def get_top_level_items(self) -> list[QW.QTreeWidgetItem]:
        """Iterate over top level items"""
        return [self.topLevelItem(_i) for _i in range(self.topLevelItemCount())]

    def find_all_items(self):
        """Find all items"""
        return self.findItems("", QC.Qt.MatchContains | QC.Qt.MatchRecursive)

    def contextMenuEvent(self, event: QG.QContextMenuEvent) -> None:
        """Override Qt method"""
        self.update_menu()
        self.menu.popup(event.globalPos())


class H5TreeWidget(AbstractTreeWidget):
    """HDF5 Browser Tree Widget

    Args:
        parent: Parent widget
    """

    SIG_SELECTED = QC.Signal(QW.QTreeWidgetItem)

    def __init__(self, parent: QW.QWidget) -> None:
        super().__init__(parent)
        title = _("HDF5 Browser")
        self.setColumnCount(4)
        self.setWindowTitle(title)
        self.setHeaderLabels([_("Name"), _("Size"), _("Type"), _("Value")])
        self.header().setSectionResizeMode(0, QW.QHeaderView.Stretch)
        self.header().setStretchLastSection(False)
        self.fnames: list[str] = []
        self.h5importers: list[H5Importer] = []

    def add_root(self, fname: str) -> None:
        """Add HDF5 root (new file)

        Args:
            fname: HDF5 file name
        """
        self.fnames.append(osp.abspath(fname))
        importer = H5Importer(fname)
        self.h5importers.append(importer)
        self.add_root_to_tree(importer)
        for col in range(1, 4):
            self.resizeColumnToContents(col)

    def remove_root(self, fname: str) -> None:
        """Remove HDF5 root

        Args:
            fname: HDF5 file name
        """
        index = self.fnames.index(osp.abspath(fname))
        self.fnames.pop(index)
        importer = self.h5importers.pop(index)
        importer.close()
        # Remove root item associated with file
        item = self.topLevelItem(index)
        self.takeTopLevelItem(index)
        del item

    def cleanup(self) -> None:
        """Clean up widget"""
        for importer in self.h5importers:
            importer.close()
        self.fnames: list[str] = []
        self.h5importers: list[H5Importer] = []
        self.clear()

    def __get_top_level_item(self, item: QW.QTreeWidgetItem) -> QW.QTreeWidgetItem:
        """Get top level item associated to item

        Args:
            item: Tree item

        Returns:
            Top level item
        """
        while item.parent():
            item = item.parent()
        return item

    def get_node(self, item: QW.QTreeWidgetItem) -> BaseNode:
        """Get HDF5 dataset associated to item

        Args:
            item: Tree item

        Returns:
            HDF5 node
        """
        toplevel_item = self.__get_top_level_item(item)
        toplevel_index = self.indexOfTopLevelItem(toplevel_item)
        node_id = item.data(0, QC.Qt.UserRole)
        if node_id:
            importer = self.h5importers[toplevel_index]
            return importer.get(node_id)
        return None

    def get_nodes(self, only_checked_items: bool = True) -> list[BaseNode]:
        """Get all nodes associated to checked items

        Args:
            only_checked_items: If True, only checked items are returned

        Returns:
            List of HDF5 nodes
        """
        datasets = []
        for item in self.find_all_items():
            if item.flags() & QC.Qt.ItemIsUserCheckable:
                if only_checked_items and item.checkState(0) == 0:
                    continue
                if item is not self.topLevelItem(0):
                    node = self.get_node(item)
                    datasets.append(node)
        return datasets

    def activated(self, item: QW.QTreeWidgetItem) -> None:
        """Double-click event"""
        if item is not self.topLevelItem(0):
            self.SIG_SELECTED.emit(item)

    def clicked(self, item: QW.QTreeWidgetItem) -> None:
        """Click event"""
        self.activated(item)

    def get_actions_from_items(self, items):  # pylint: disable=W0613
        """Get actions from item"""
        return []

    def is_empty(self) -> bool:
        """Return True if tree is empty"""
        return len(self.find_all_items()) == 1

    def is_any_item_checked(self) -> bool:
        """Return True if any item is checked"""
        for item in self.find_all_items():
            if item.checkState(0) > 0:
                return True
        return False

    def select_all(self, state: bool) -> None:
        """Select all items

        Args:
            state: If True, all items are selected
        """
        for item in self.find_all_items():
            if item.flags() & QC.Qt.ItemIsUserCheckable:
                item.setSelected(state)
                if state:
                    self.clicked(item)

    def toggle_all(self, state: bool) -> None:
        """Toggle all item state from 'unchecked' to 'checked'
        (or vice-versa)

        Args:
            state: If True, all items are checked
        """
        for item in self.find_all_items():
            if item.flags() & QC.Qt.ItemIsUserCheckable:
                item.setCheckState(0, QC.Qt.Checked if state else QC.Qt.Unchecked)

    @staticmethod
    def __create_node(node: BaseNode) -> QW.QTreeWidgetItem:
        """Create tree node from HDF5 node

        Args:
            node: HDF5 node

        Returns:
            Tree widget node
        """
        text = to_string(node.text)
        if len(text) > 30:
            text = text[:30] + "..."
        treeitem = QW.QTreeWidgetItem([node.name, node.shape_str, node.dtype_str, text])
        treeitem.setData(0, QC.Qt.UserRole, node.id)
        if node.description:
            for col in range(treeitem.columnCount()):
                treeitem.setToolTip(col, node.description)
        return treeitem

    @staticmethod
    def __recursive_popfunc(parent_item: QW.QTreeWidgetItem, node: BaseNode) -> None:
        """Recursive HDF5 analysis

        Args:
            parent_item: Parent tree item
            node: HDF5 node
        """
        tree_item = H5TreeWidget.__create_node(node)
        if node.is_supported():
            tree_item.setCheckState(0, QC.Qt.Unchecked)
        else:
            tree_item.setFlags(QC.Qt.ItemIsEnabled)
        tree_item.setIcon(0, get_icon(node.icon_name))
        parent_item.addChild(tree_item)
        for child in node.children:
            H5TreeWidget.__recursive_popfunc(tree_item, child)

    def expand_all_children(self, item: QW.QTreeWidgetItem) -> None:
        """Expand all children (recursively)

        Args:
            item: Tree item
        """
        self.expandItem(item)
        for index in range(item.childCount()):
            child = item.child(index)
            self.expand_all_children(child)

    def add_root_to_tree(self, importer: H5Importer) -> None:
        """Add root to tree

        Args:
            importer: HDF5 importer
        """
        root = importer.root
        rootitem = QW.QTreeWidgetItem([root.name])
        rootitem.setToolTip(0, root.description)
        rootitem.setData(0, QC.Qt.UserRole, root.id)
        rootitem.setFlags(QC.Qt.ItemIsEnabled)
        rootitem.setIcon(0, get_icon(root.icon_name))
        self.addTopLevelItem(rootitem)
        for node in root.children:
            self.__recursive_popfunc(rootitem, node)
        self.expand_all_children(rootitem)

    def toggle_show_only_checkable_items(self, state: bool) -> None:
        """Show only checkable items

        Args:
            state: If True, only checkable items are shown
        """
        for item in self.find_all_items():
            item.setHidden(state)
        if state:
            # Iterate over checkable items and show them (and their parents)
            for item in self.find_all_items():
                if item.flags() & QC.Qt.ItemIsUserCheckable:
                    item.setHidden(False)
                    parent = item.parent()
                    while parent:
                        parent.setHidden(False)
                        parent = parent.parent()

    def toggle_show_values(self, state: bool) -> None:
        """Show values

        Args:
            state: If True, values are shown
        """
        # Hide or show the "Value" column
        self.setColumnHidden(3, not state)

    def set_current_file(self, fname: str) -> None:
        """Set current file

        Args:
            fname: HDF5 file name
        """
        index = self.fnames.index(osp.abspath(fname))
        item = self.topLevelItem(index)
        self.setCurrentItem(item)
        self.scrollToItem(item, QW.QAbstractItemView.PositionAtTop)


class PlotPreview(QW.QStackedWidget):
    """Plot preview"""

    def __init__(self, parent: QW.QWidget) -> None:
        super().__init__(parent)
        self.curvewidget = PlotWidget(
            self, options=PlotOptions(type="curve", curve_antialiasing=True)
        )
        self.addWidget(self.curvewidget)
        self.imagewidget = PlotWidget(
            self, options=PlotOptions(type="image", show_contrast=True)
        )
        self.addWidget(self.imagewidget)

    def cleanup(self) -> None:
        """Clean up widget"""
        for widget in (self.imagewidget, self.curvewidget):
            widget.get_plot().del_all_items()

    def update_plot_preview(self, node: BaseNode) -> None:
        """Update plot preview widget"""
        try:
            obj = node.get_native_object()
        except Exception as msg:  # pylint: disable=broad-except
            qt_handle_error_message(self, msg)
            return
        if obj is None:
            # An error occurred while creating the object (invalid data, ...)
            label = make.label(_("Unsupported data"), "C", (0, 0), "C")
            plot: BasePlot = self.currentWidget().get_plot()
            plot.del_all_items()
            plot.add_item(label)
            plot.replot()
            return
        if isinstance(obj, SignalObj):
            obj: SignalObj
            widget = self.curvewidget
        else:
            obj: ImageObj
            widget = self.imagewidget
        with CURVESTYLES.suspend():
            item = obj.make_item()
        plot = widget.get_plot()
        plot.del_all_items()
        plot.add_item(item)
        plot.set_active_item(item)
        item.unselect()
        plot.do_autoscale()
        self.setCurrentWidget(widget)


class TablePreview(QW.QWidget):
    """Table preview

    Args:
        title: Group title
        parent: Parent widget
    """

    def __init__(self, parent: QW.QWidget) -> None:
        super().__init__(parent)
        self.setLayout(QW.QVBoxLayout())
        self.table = QW.QTableWidget(self)
        self.table.setEditTriggers(QW.QAbstractItemView.NoEditTriggers)
        self.table.horizontalHeader().setStretchLastSection(True)
        self.layout().addWidget(self.table)

    def clear(self) -> None:
        """Clear table"""
        self.table.clear()

    def update_table_preview(self, data: dict[str, Any]) -> None:
        """Update table preview widget

        Args:
            node: HDF5 node
        """
        self.clear()
        self.table.setRowCount(len(data))
        self.table.setColumnCount(1)
        self.table.setHorizontalHeaderLabels([_("Value")])
        self.table.setVerticalHeaderLabels(list(data.keys()))
        for row, value in enumerate(data.values()):
            self.table.setItem(row, 0, QW.QTableWidgetItem(str(value)))
        self.table.resizeRowsToContents()


class GroupAndAttributes(QW.QTabWidget):
    """Group and attributes

    Args:
        parent: Parent widget
        show_array_callback: Callback to show array
    """

    def __init__(self, parent: QW.QWidget, show_array_callback: Callable) -> None:
        super().__init__(parent)
        self.group = TablePreview(self)
        self.addTab(self.group, get_icon("h5group.svg"), _("Group"))
        self.attrs = TablePreview(self)
        self.addTab(self.attrs, get_icon("h5attrs.svg"), _("Attributes"))
        # Add a button as corner widget to show the array (if any):
        self.__show_array_btn = create_toolbutton(
            self,
            icon=get_icon("show_results.svg"),
            text=_("Show array"),
            autoraise=False,
            triggered=show_array_callback,
        )
        self.__show_array_btn.setEnabled(False)
        self.setCornerWidget(self.__show_array_btn, QC.Qt.TopRightCorner)

    def cleanup(self) -> None:
        """Clean up widget"""
        self.group.clear()
        self.attrs.clear()

    def update_from_node(self, node: BaseNode) -> None:
        """Update widget from node

        Args:
            node: HDF5 node
        """
        # Update group =================================================================
        text = to_string(node.text)
        if text:
            lines = text.splitlines()[:5]
            if len(lines) == 5:
                lines += ["[...]"]
            text = os.linesep.join(lines)
        data = {
            _("Path"): node.id,
            _("Name"): node.name,
            _("Description"): node.description,
            _("Textual preview"): text,
            # "Raw": repr(node.data),
        }
        self.group.update_table_preview(data)

        # Update attributes ============================================================
        self.attrs.update_table_preview(node.metadata)

        # Update show array button =====================================================
        self.__show_array_btn.setEnabled(node.IS_ARRAY)


class H5FileSelector(QW.QWidget):
    """HDF5 file selector

    Args:
        parent: Parent widget
    """

    SIG_ADD_FILENAME = QC.Signal(str)
    SIG_REMOVE_FILENAME = QC.Signal(str)
    SIG_CURRENT_CHANGED = QC.Signal(str)

    def __init__(self, parent: QW.QWidget) -> None:
        super().__init__(parent)
        self.setLayout(QW.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.combo = QW.QComboBox(self)
        self.combo.currentTextChanged.connect(self.current_file_changed)
        self.layout().addWidget(self.combo)
        self.btn_add = create_toolbutton(
            self,
            icon=get_std_icon("DirOpenIcon"),
            text=_("Open") + " ...",
            autoraise=False,
            triggered=lambda _checked=False: self.add_file(),
        )
        self.btn_add.setSizePolicy(QW.QSizePolicy.Fixed, QW.QSizePolicy.Fixed)
        self.layout().addWidget(self.btn_add)
        self.btn_rmv = create_toolbutton(
            self,
            icon=get_std_icon("DialogCloseButton"),
            text=_("Close"),
            autoraise=False,
            triggered=self.remove_file,
        )
        self.btn_rmv.setSizePolicy(QW.QSizePolicy.Fixed, QW.QSizePolicy.Fixed)
        self.layout().addWidget(self.btn_rmv)
        self.btn_rmv.setEnabled(False)

    def set_current_fname(self, fname: str) -> None:
        """Set current file name

        Args:
            fname: HDF5 file name
        """
        index = self.combo.findText(fname)
        if index >= 0:
            self.combo.setCurrentIndex(index)

    def get_current_fname(self) -> str:
        """Return current file name

        Returns:
            HDF5 file name
        """
        return self.combo.currentText()

    def current_file_changed(self, fname: str) -> None:
        """Current file changed

        Args:
            fname: HDF5 file name
        """
        self.SIG_CURRENT_CHANGED.emit(fname)

    def add_fname(self, fname: str) -> None:
        """Add file name

        Args:
            fname: HDF5 file name
        """
        self.combo.addItem(get_icon("h5file.svg"), fname)
        self.btn_rmv.setEnabled(True)

    def remove_fname(self, fname: str) -> None:
        """Remove file name

        Args:
            fname: HDF5 file name
        """
        index = self.combo.findText(fname)
        if index >= 0:
            self.combo.removeItem(index)
        if self.combo.count() == 0:
            self.btn_rmv.setEnabled(False)

    def add_file(self, fname: str | None = None) -> None:
        """Browse file

        Args:
            fname: HDF5 file name. Default is None.
             (this is used for testing only)
        """
        if fname is None:
            fname = getopenfilename(
                self, _("Select HDF5 file"), "", _("HDF5 files (*.h5 *.hdf5)")
            )[0]
        if fname:
            self.SIG_ADD_FILENAME.emit(osp.abspath(fname))

    def remove_file(self, fname: str | None = None) -> None:
        """Remove file name

        Args:
            fname: HDF5 file name
        """
        if fname is None:
            fname = self.combo.currentText()
        self.SIG_REMOVE_FILENAME.emit(fname)


class H5Browser(QW.QSplitter):
    """HDF5 Browser Widget

    Args:
        parent: Parent widget
    """

    SIG_SELECT_NEW_FILE = QC.Signal(str)
    SIG_REMOVE_FILE = QC.Signal(str)

    def __init__(self, parent: QW.QWidget | None = None) -> None:
        super().__init__(parent)
        self.selector = H5FileSelector(self)
        self.selector.SIG_ADD_FILENAME.connect(self.__add_new_file)
        self.selector.SIG_REMOVE_FILENAME.connect(self.__remove_file)
        self.selector.SIG_CURRENT_CHANGED.connect(self.__selector_current_file_changed)
        self.tree = H5TreeWidget(self)
        self.tree.SIG_SELECTED.connect(self.__item_selected_on_tree)
        selectorandtree = QW.QFrame(self)
        selectorandtree.setLayout(QW.QVBoxLayout())
        selectorandtree.layout().addWidget(self.selector)
        selectorandtree.layout().addWidget(self.tree)
        selectorandtree.layout().setContentsMargins(0, 0, 0, 0)
        self.addWidget(selectorandtree)
        preview = QW.QSplitter(self)
        preview.setOrientation(QC.Qt.Vertical)
        self.addWidget(preview)
        self.plotpreview = PlotPreview(self)
        preview.addWidget(self.plotpreview)
        self.groupandattrs = GroupAndAttributes(self, self.show_array)
        preview.addWidget(self.groupandattrs)
        preview.setSizes([int(self.size().height() / 2)] * 2)

    def open_file(self, fname: str) -> None:
        """Open HDF5 file

        Args:
            fname: HDF5 file name
        """
        self.tree.add_root(fname)
        self.selector.add_fname(fname)

    def close_file(self, fname: str) -> None:
        """Close HDF5 file

        Args:
            fname: HDF5 file name
        """
        self.tree.remove_root(fname)
        self.selector.remove_fname(fname)

    def __add_new_file(self, fname: str) -> None:
        """Add new file

        Args:
            fname: HDF5 file name
        """
        self.open_file(fname)
        self.selector.set_current_fname(fname)
        self.SIG_SELECT_NEW_FILE.emit(fname)

    def __remove_file(self, fname: str) -> None:
        """Remove file

        Args:
            fname: HDF5 file name
        """
        self.close_file(fname)
        self.SIG_REMOVE_FILE.emit(fname)

    def cleanup(self) -> None:
        """Clean up widget"""
        self.tree.cleanup()
        self.plotpreview.cleanup()

    def get_node(self, item: QW.QTreeWidgetItem | None = None) -> BaseNode:
        """Return (selected) dataset

        Args:
            item: Tree item

        Returns:
            HDF5 node
        """
        if item is None:
            item = self.tree.currentItem()
        return self.tree.get_node(item)

    def __item_selected_on_tree(self, item: QW.QTreeWidgetItem) -> None:
        """Item selected on tree

        Args:
            item: Tree item
        """
        # View the selected item
        node = self.get_node(item)
        if node.is_supported():
            self.plotpreview.update_plot_preview(node)
        self.groupandattrs.update_from_node(node)
        # Update the file selector combo box
        self.selector.set_current_fname(node.h5file.filename)

    def __selector_current_file_changed(self, fname: str) -> None:
        """Selector current file changed

        Args:
            fname: HDF5 file name
        """
        if fname:
            self.tree.set_current_file(fname)

    def show_array(self) -> None:
        """Show array"""
        node = self.get_node()
        assert node.IS_ARRAY
        arrayeditor = ArrayEditor(self)
        arrayeditor.setup_and_check(node.data, title=node.name, readonly=True)
        exec_dialog(arrayeditor)


class H5BrowserDialog(QW.QDialog):
    """HDF5 Browser Dialog

    Args:
        parent: Parent widget
        size: Dialog size
    """

    def __init__(
        self, parent: QW.QWidget | None = None, size: tuple[int, int] = (1150, 700)
    ) -> None:
        super().__init__(parent)
        self.setWindowFlags(QC.Qt.Window)
        self.setObjectName("h5browser")
        self.setWindowTitle(_("HDF5 Browser"))
        self.setWindowIcon(get_icon("h5browser.svg"))
        win32_fix_title_bar_background(self)
        vlayout = QW.QVBoxLayout()
        self.setLayout(vlayout)
        self.button_layout: QW.QHBoxLayout | None = None
        self.bbox: QW.QDialogButtonBox | None = None
        self.nodes: list[BaseNode] = []
        self.checkbox_show_only: QW.QCheckBox | None = None
        self.checkbox_show_values: QW.QCheckBox | None = None

        self.browser = H5Browser(self)
        self.browser.SIG_SELECT_NEW_FILE.connect(self.select_new_file)
        self.browser.SIG_REMOVE_FILE.connect(self.remove_file)
        vlayout.addWidget(self.browser)

        self.browser.tree.itemChanged.connect(lambda item: self.refresh_buttons())

        self.install_button_layout()

        self.setMinimumSize(QC.QSize(900, 500))
        self.resize(QC.QSize(*size))
        self.browser.setSizes([int(self.size().height() / 2)] * 2)
        self.refresh_buttons()

    def accept(self) -> None:
        """Accept changes"""
        self.nodes = self.browser.tree.get_nodes()
        QW.QDialog.accept(self)

    def is_empty(self) -> bool:
        """Return True if tree is empty"""
        return self.browser.tree.is_empty()

    def cleanup(self) -> None:
        """Cleanup dialog"""
        self.browser.cleanup()

    def refresh_buttons(self) -> None:
        """Refresh buttons"""
        state = self.browser.tree.is_any_item_checked()
        self.bbox.button(QW.QDialogButtonBox.Ok).setEnabled(state)

    def show_only_checkable_items(self, state: int) -> None:
        """Show only checkable items

        Args:
            state: If True, only checkable items are shown
        """
        self.browser.tree.toggle_show_only_checkable_items(state)
        fname = self.browser.selector.get_current_fname()
        if fname:
            self.browser.tree.set_current_file(fname)

    def __finalize_setup(self) -> None:
        """Finalize setup"""
        tree = self.browser.tree
        tree.toggle_show_only_checkable_items(self.checkbox_show_only.isChecked())
        tree.toggle_show_values(self.checkbox_show_values.isChecked())

    def open_file(self, fname: str) -> None:
        """Open file

        Args:
            fname: HDF5 file name
        """
        self.browser.open_file(fname)
        self.__finalize_setup()

    def open_files(self, fnames: list[str]) -> None:
        """Open files

        Args:
            fnames: HDF5 file names
        """
        for fname in fnames:
            self.browser.open_file(fname)
        self.__finalize_setup()

    def select_new_file(self, fname: str) -> None:  # pylint:disable=unused-argument
        """Select new file

        Args:
            fname: HDF5 file name
        """
        self.__finalize_setup()
        self.refresh_buttons()

    def remove_file(self, fname: str) -> None:  # pylint:disable=unused-argument
        """Remove file

        Args:
            fname: HDF5 file name
        """
        self.refresh_buttons()

    def get_all_nodes(self) -> list[BaseNode]:
        """Return all supported datasets

        Returns:
            List of HDF5 nodes
        """
        return self.browser.tree.get_nodes(only_checked_items=False)

    def get_nodes(self) -> list[BaseNode]:
        """Return datasets

        Returns:
            List of HDF5 nodes
        """
        return self.nodes

    def install_button_layout(self) -> None:
        """Install button layout"""
        bbox = QW.QDialogButtonBox(QW.QDialogButtonBox.Ok | QW.QDialogButtonBox.Cancel)
        bbox.accepted.connect(self.accept)
        bbox.rejected.connect(self.reject)

        btn_check_all = create_toolbutton(
            self,
            icon=get_icon("check_all.svg"),
            text=_("Check all"),
            autoraise=False,
            shortcut=QG.QKeySequence.SelectAll,
            triggered=lambda checked=True: self.browser.tree.toggle_all(checked),
        )
        btn_uncheck_all = create_toolbutton(
            self,
            icon=get_icon("uncheck_all.svg"),
            text=_("Uncheck all"),
            autoraise=False,
            triggered=lambda checked=False: self.browser.tree.toggle_all(checked),
        )
        self.checkbox_show_only = QW.QCheckBox(_("Show only supported data"))
        self.checkbox_show_only.stateChanged.connect(self.show_only_checkable_items)
        self.checkbox_show_values = QW.QCheckBox(_("Show values"))
        self.checkbox_show_values.stateChanged.connect(
            self.browser.tree.toggle_show_values
        )

        self.button_layout = QW.QHBoxLayout()
        self.button_layout.addWidget(self.checkbox_show_only)
        self.button_layout.addWidget(self.checkbox_show_values)
        self.button_layout.addSpacing(10)
        self.button_layout.addWidget(btn_check_all)
        self.button_layout.addWidget(btn_uncheck_all)
        self.button_layout.addStretch()
        self.button_layout.addWidget(bbox)
        self.bbox = bbox

        vlayout: QW.QVBoxLayout = self.layout()
        vlayout.addSpacing(10)
        vlayout.addLayout(self.button_layout)



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab Configuration utilities
"""

from __future__ import annotations

import os
import os.path as osp
import warnings
from typing import Any

from guidata.userconfig import NoDefault, UserConfig


class AppUserConfig(UserConfig):
    """Application user configuration"""

    def to_dict(self) -> dict:
        """Return configuration as a dictionary"""
        confdict = {}
        for section in self.sections():
            secdict = {}
            for option, value in self.items(section, raw=self.raw):
                secdict[option] = value
            confdict[section] = secdict
        return confdict


CONF = AppUserConfig({})


class Configuration:
    """Configuration file"""

    @classmethod
    def initialize(cls, name: str, version: str, load: bool) -> None:
        """Initialize configuration"""
        CONF.set_application(name, version, load=load)

    @classmethod
    def reset(cls) -> None:
        """Reset configuration"""
        global CONF  # pylint: disable=global-statement
        CONF.cleanup()  # Remove configuration file
        CONF = AppUserConfig({})

    @classmethod
    def get_filename(cls) -> str:
        """Return configuration file name"""
        return CONF.filename()

    @classmethod
    def get_path(cls, basename: str) -> str:
        """Return filename path inside configuration directory"""
        return CONF.get_path(basename)

    @classmethod
    def to_dict(cls) -> dict:
        """Return configuration as a dictionary"""
        return CONF.to_dict()


class Section:
    """Configuration section"""

    @classmethod
    def set_name(cls, section: str) -> None:
        """Set section name"""
        cls._name = section

    @classmethod
    def get_name(cls) -> str:
        """Return section name"""
        return cls._name


class Option:
    """Configuration option handler"""

    def __init__(self) -> None:
        self.section = None
        self.option = None

    def get(self, default=NoDefault) -> Any:
        """Get configuration option value"""
        return CONF.get(self.section, self.option, default)

    def set(self, value: Any) -> None:
        """Set configuration option value"""
        CONF.set(self.section, self.option, value)

    def remove(self) -> None:
        """Remove configuration option"""
        # No use case for this method yet (quite dangerous!)
        CONF.remove_option(self.section, self.option)


class ConfigPathOption(Option):
    """Configuration file path configuration option handler"""

    def get(self, default=NoDefault) -> str:
        """Get configuration file path from configuration"""
        if default is NoDefault:
            default = ""
        fname = super().get(default)
        if osp.basename(fname) != fname:
            raise ValueError(f"Invalid configuration file name {fname}")
        return CONF.get_path(osp.basename(fname))


class WorkingDirOption(Option):
    """Working directory configuration option handler"""

    def get(self, default=NoDefault) -> str:
        """Get working directory from configuration"""
        if default is NoDefault:
            default = ""
        path = super().get(default)
        if osp.isdir(path):
            return path
        return ""

    def set(self, value: str) -> None:
        """Set working directory in configuration"""
        if not osp.isdir(value):
            value = osp.dirname(value)
            if not osp.isdir(value):
                raise FileNotFoundError(f"Invalid working directory name {value}")
        os.chdir(value)
        super().set(value)


class EnumOption(Option):
    """Enumeration option handler"""

    def __init__(self, values: list[Any], default: Any = NoDefault) -> None:
        super().__init__()
        if default is NoDefault:
            default = values[0]
        self.values = values
        self.default = default

    def get(self, default: Any = NoDefault) -> Any:
        """Get configuration option value"""
        value = super().get(default)
        if value not in self.values:
            # Only show a warning here, as the configuration file may be edited manually
            warnings.warn(
                f"Invalid value {value} for option {self.option}, "
                f"expected {self.values}"
            )
            return self.default
        return value

    def set(self, value: Any) -> None:
        """Set configuration option value"""
        if value not in self.values:
            raise ValueError(
                f"Invalid value {value} for option {self.option}, "
                f"expected {self.values}"
            )
        super().set(value)


class SectionMeta(type):
    """Configuration metaclass"""

    #  pylint: disable=bad-mcs-classmethod-argument
    def __new__(cls, name, bases, dct):
        optlist = []
        for attrname, obj in list(dct.items()):
            if isinstance(obj, Option):
                obj.option = attrname
                optlist.append(obj)
        dct["_options"] = optlist
        return type.__new__(cls, name, bases, dct)


class ConfMeta(type):
    """Configuration metaclass"""

    #  pylint: disable=bad-mcs-classmethod-argument
    def __new__(cls, name, bases, dct):
        for attrname, obj in list(dct.items()):
            if isinstance(obj, Section):
                obj.set_name(attrname)
                for option in obj._options:
                    option.section = attrname
        return type.__new__(cls, name, bases, dct)



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Tests (:mod:`cdl.tests`)
------------------------

The DataLab test suite is based on the `pytest <https://pytest.org>`_ framework.

The test suite modules are organized in subpackages according to their purpose.
The following subpackages are available:

- :mod:`cdl.tests.backbone`: backbone tests
- :mod:`cdl.tests.features`: feature tests (unit tests and application tests)
- :mod:`cdl.tests.scenarios`: high-level scenarios tests

.. seealso::

    :ref:`validation` for more information about DataLab's testing strategy.
"""

from __future__ import annotations

import os
from collections.abc import Generator
from contextlib import contextmanager

from guidata.guitest import run_testlauncher

import cdl.config  # Loading icons
from cdl.core.gui.main import CDLMainWindow
from cdl.core.gui.panel.image import ImagePanel
from cdl.core.gui.panel.signal import SignalPanel
from cdl.env import execenv
from cdl.utils import qthelpers as qth
from cdl.utils import tests


@contextmanager
def cdltest_app_context(
    size: tuple[int, int] = None,
    maximized: bool = False,
    save: bool = False,
    console: bool | None = None,
    exec_loop: bool = True,
) -> Generator[CDLMainWindow, None, None]:
    """Context manager handling DataLab mainwindow creation and Qt event loop
    with optional HDF5 file save and other options for testing purposes

    Args:
        size: mainwindow size (default: (950, 600))
        maximized: whether to maximize mainwindow (default: False)
        save: whether to save HDF5 file (default: False)
        console: whether to show console (default: None)
        exec_loop: whether to execute Qt event loop (default: True)
    """
    if size is None:
        size = 950, 600
    with qth.cdl_app_context(exec_loop=exec_loop):
        try:
            win = CDLMainWindow(console=console)
            if maximized:
                win.showMaximized()
            else:
                width, height = size
                win.resize(width, height)
                win.showNormal()
            win.show()
            win.setObjectName(tests.get_default_test_name())  # screenshot name
            yield win
        finally:
            if save:
                path = tests.get_output_data_path("h5")
                try:
                    os.remove(path)
                    win.save_to_h5_file(path)
                except (FileNotFoundError, PermissionError):
                    pass
            if not exec_loop:
                # Closing main window properly
                win.set_modified(False)
                win.close()


def take_plotwidget_screenshot(panel: SignalPanel | ImagePanel, name: str) -> None:
    """Eventually takes plotwidget screenshot (only in screenshot mode)"""
    if execenv.screenshot:
        prefix = panel.PARAMCLASS.PREFIX
        qth.grab_save_window(panel.plothandler.plotwidget, f"{prefix}_{name}")


def run() -> None:
    """Run DataLab test launcher"""
    run_testlauncher(cdl)


if __name__ == "__main__":
    run()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Image centroid computation test

Comparing different algorithms for centroid calculation:

- SciPy (measurements.center_of_mass)
- OpenCV (moments)
- Method based on moments
- Method based on Fourier (DataLab)
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code
# guitest: show

import time

import numpy as np
import pytest
import scipy.ndimage as spi
from guidata.qthelpers import qt_app_context
from numpy import ma
from plotpy.builder import make

import cdl.computation.image as cpi
import cdl.obj as dlo
from cdl.algorithms.image import get_centroid_fourier
from cdl.config import _
from cdl.env import execenv
from cdl.tests.data import (
    create_noisygauss_image,
    get_laser_spot_data,
)
from cdl.utils.tests import check_scalar_result
from cdl.utils.vistools import view_image_items


def get_centroid_from_moments(data):
    """Computing centroid from image moments"""
    y, x = np.ogrid[: data.shape[0], : data.shape[1]]
    imx, imy = data.sum(axis=0)[None, :], data.sum(axis=1)[:, None]
    m00 = np.array(data, dtype=float).sum() or 1.0
    m10 = (np.array(imx, dtype=float) * x).sum() / m00
    m01 = (np.array(imy, dtype=float) * y).sum() / m00
    return int(m01), int(m10)


def get_centroid_with_cv2(data):
    """Compute centroid from moments with OpenCV"""
    import cv2  # pylint: disable=import-outside-toplevel

    m = cv2.moments(data)
    col = int(m["m10"] / m["m00"])
    row = int(m["m01"] / m["m00"])
    return row, col


def add_xcursor(items, x, y, title):
    """Added X cursor to plot"""
    label = "  " + f'{_("Centroid")}[{title}] (x=%s, y=%s)'
    execenv.print(label % (x, y))
    cursor = make.xcursor(x, y, label=label)
    cursor.setTitle(title)
    items.append(cursor)


def compare_centroid_funcs(data):
    """Compare centroid methods"""
    items = []
    items += [make.image(data, interpolation="nearest", eliminate_outliers=2.0)]
    # Computing centroid coordinates
    for name, func in (
        ("SciPy", spi.center_of_mass),
        ("OpenCV", get_centroid_with_cv2),
        ("Moments", get_centroid_from_moments),
        ("Fourier", get_centroid_fourier),
    ):
        try:
            t0 = time.time()
            row, col = func(data)
            dt = time.time() - t0
            add_xcursor(items, col, row, name)
            execenv.print(f"    Calculation time: {int(dt * 1e3):d} ms")
        except ImportError:
            execenv.print(f"    Unable to compute {name}: missing module")
    view_image_items(items)


def test_centroid_graphically():
    """Centroid test comparing different methods and showing results"""
    with qt_app_context():
        for data in get_laser_spot_data():
            execenv.print(f"Data[dtype={data.dtype},shape={data.shape}]")
            # Testing with masked arrays
            compare_centroid_funcs(data.view(ma.MaskedArray))


def __check_centroid(image, expected_x, expected_y):
    """Check centroid computation"""
    df = cpi.compute_centroid(image).to_dataframe()
    check_scalar_result("Centroid X", df.x[0], expected_x, atol=1.0)
    check_scalar_result("Centroid Y", df.y[0], expected_y, atol=1.0)


@pytest.mark.validation
def test_image_centroid():
    """Test centroid computation"""
    param = dlo.NewImageParam.create(height=500, width=500)
    image = create_noisygauss_image(param, center=(-2.0, 3.0), add_annotations=True)
    circle_roi = dlo.create_image_roi("circle", [200, 325, 10])
    for roi, x0, y0 in (
        (None, 0, 0),
        (None, 100, 100),
        (circle_roi, 0, 0),
        (circle_roi, 100, 100),  # Test for regression like #106
    ):
        image.roi, image.x0, image.y0 = roi, x0, y0
        __check_centroid(image, 200.0 + x0, 325.0 + y0)


if __name__ == "__main__":
    test_centroid_graphically()
    test_image_centroid()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
New signal/image test

Testing functions related to signal/image creation.
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code
# guitest: show

from __future__ import annotations

from collections.abc import Generator

from guidata.qthelpers import qt_app_context

from cdl.env import execenv
from cdl.obj import (
    Gauss2DParam,
    ImageDatatypes,
    ImageObj,
    ImageTypes,
    NormalRandomParam,
    SignalObj,
    SignalTypes,
    UniformRandomParam,
    create_image_from_param,
    create_signal_from_param,
    new_image_param,
    new_signal_param,
)
from cdl.utils.vistools import view_curves, view_images


def iterate_signal_creation(
    data_size: int = 500, non_zero: bool = False, verbose: bool = True
) -> Generator[SignalObj, None, None]:
    """Iterate over all possible signals created from parameters"""
    if verbose:
        execenv.print(
            f"  Iterating over signal types (size={data_size}, non_zero={non_zero}):"
        )
    for stype in SignalTypes:
        if non_zero and stype in (SignalTypes.ZEROS,):
            continue
        if verbose:
            execenv.print(f"    {stype.value}")
        newparam = new_signal_param(stype=stype, size=data_size)
        if stype == SignalTypes.UNIFORMRANDOM:
            addparam = UniformRandomParam()
        elif stype == SignalTypes.NORMALRANDOM:
            addparam = NormalRandomParam()
        else:
            addparam = None
        signal = create_signal_from_param(newparam, addparam=addparam)
        if stype == SignalTypes.ZEROS:
            assert (signal.y == 0).all()
        yield signal


def iterate_image_creation(
    data_size: int = 500, non_zero: bool = False, verbose: bool = True
) -> Generator[ImageObj, None, None]:
    """Iterate over all possible images created from parameters"""
    if verbose:
        execenv.print(
            f"  Iterating over image types (size={data_size}, non_zero={non_zero}):"
        )
    for itype in ImageTypes:
        if non_zero and itype in (ImageTypes.EMPTY, ImageTypes.ZEROS):
            continue
        if verbose:
            execenv.print(f"    {itype.value}")
        for dtype in ImageDatatypes:
            if verbose:
                execenv.print(f"      {dtype.value}")
            newparam = new_image_param(
                itype=itype, dtype=dtype, width=data_size, height=data_size
            )
            if itype == ImageTypes.GAUSS:
                addparam = Gauss2DParam()
                addparam.x0 = addparam.y0 = 3
                addparam.sigma = 5
            elif itype == ImageTypes.UNIFORMRANDOM:
                addparam = UniformRandomParam()
                addparam.set_from_datatype(dtype.value)
            elif itype == ImageTypes.NORMALRANDOM:
                addparam = NormalRandomParam()
                addparam.set_from_datatype(dtype.value)
            else:
                addparam = None
            image = create_image_from_param(newparam, addparam=addparam)
            if itype == ImageTypes.ZEROS:
                assert (image.data == 0).all()
            yield image


def all_combinations_test():
    """Test all combinations for new signal/image feature"""
    execenv.print(f"Testing {all_combinations_test.__name__}:")
    execenv.print(f"  Signal types ({len(SignalTypes)}):")
    for signal in iterate_signal_creation():
        assert signal.x is not None and signal.y is not None
    execenv.print(f"  Image types ({len(ImageTypes)}):")
    for image in iterate_image_creation():
        assert image.data is not None
    execenv.print(f"{all_combinations_test.__name__} OK")


def __new_signal_test():
    """Test new signal feature"""
    edit = not execenv.unattended
    signal = create_signal_from_param(None, edit=edit)
    if signal is not None:
        data = (signal.x, signal.y)
        view_curves([data], name=__new_signal_test.__name__, title=signal.title)


def __new_image_test():
    """Test new image feature"""
    # Test with no input parameter
    edit = not execenv.unattended
    image = create_image_from_param(None, edit=edit)
    if image is not None:
        view_images(image.data, name=__new_image_test.__name__, title=image.title)
    # Test with parametered 2D-Gaussian
    newparam = new_image_param(itype=ImageTypes.GAUSS)
    addparam = Gauss2DParam()
    addparam.x0 = addparam.y0 = 3
    addparam.sigma = 5
    image = create_image_from_param(newparam, addparam=addparam, edit=edit)
    if image is not None:
        view_images(image.data, name=__new_image_test.__name__, title=image.title)


def test_new_object():
    """Test new signal/image feature"""
    all_combinations_test()
    with qt_app_context():
        __new_signal_test()
        __new_image_test()


if __name__ == "__main__":
    test_new_object()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Signal base line selection unit test.
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# guitest: show

import numpy as np
from guidata.qthelpers import exec_dialog, qt_app_context

from cdl.env import execenv
from cdl.tests.data import create_paracetamol_signal
from cdl.widgets.signalbaseline import SignalBaselineDialog


def test_signal_baseline_selection():
    """Signal baseline selection dialog test"""
    sig = create_paracetamol_signal()
    with qt_app_context():
        dlg = SignalBaselineDialog(sig)
        dlg.resize(640, 480)
        dlg.setObjectName(dlg.objectName() + "_00")  # to avoid timestamp suffix
        exec_dialog(dlg)
    execenv.print(f"baseline: {dlg.get_baseline()}")
    execenv.print(f"X range: {dlg.get_x_range()}")
    # Check baseline value:
    i0, i1 = np.searchsorted(sig.x, dlg.get_x_range())
    assert dlg.get_baseline() == sig.data[i0:i1].mean()


if __name__ == "__main__":
    test_signal_baseline_selection()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

"""
Tour
====

The :mod:`cdl.core.gui.tour` module implements a tour of DataLab features.

A dialog box is displayed when the user launches DataLab for the first time or
when the user clicks on the "Show tour" entry in the "?" menu.

The tour user experience is the following:

- First, the DataLab main window is maximized, and it is grayed out (the whole window
  is covered with a gray area with an opacity of 50%). The user can not interact with
  the DataLab main window while the tour is running.

- A first modal dialog box is displayed with a short description of DataLab, and two
  buttons: "Start" and "Close". If the user clicks on "Start", the tour
  starts. If the user clicks on "Close", the tour is not started and the dialog box
  is closed.

- The tour is composed of several steps. Each step is a modal dialog box with a short
  description of a DataLab feature, and three buttons: "Previous", "Next" and "Close".
  If the user clicks on "Previous", the previous step is displayed. If the user clicks
  on "Next", the next step is displayed. If the user clicks on "Close", the tour is
  stopped and the dialog box is closed.

- At each step, the dialog box is moved to the center of the DataLab main window (i.e.
  the center of the screen), and the feature described in the step is highlighted in
  the DataLab main window. The highlight is a red rectangle around the feature, and
  the feature is the only part of the DataLab main window that is not grayed out.

- The last step of the tour is a modal dialog box with a conclusion message and two
  buttons: "Show tour again", "Close", and "Show demo". If the user clicks on "Show
  tour again", the tour starts again. If the user clicks on "Close", the tour is
  stopped and the dialog box is closed. If the user clicks on "Show demo", the demo
  is started.

.. autoclass:: Tour
    :members:
    :inherited-members:
"""

from __future__ import annotations

import abc
import enum
import os
from collections.abc import Callable
from typing import TYPE_CHECKING

import numpy as np
from guidata.configtools import get_icon, get_image_file_path
from guidata.qthelpers import is_dark_theme
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qtpy import QtWidgets as QW

import cdl.obj
from cdl.config import _
from cdl.env import execenv

if TYPE_CHECKING:
    from cdl.core.gui.main import CDLMainWindow


class Cover(QW.QWidget):
    """
    Widget that covers the main window with a gray area with an opacity of 50%.
    """

    def __init__(self, parent: CDLMainWindow) -> None:
        super().__init__(parent)
        self.setWindowFlags(QC.Qt.Tool | QC.Qt.FramelessWindowHint)
        self.setAttribute(QC.Qt.WA_TranslucentBackground)
        self.__opacity_factor = 1.0
        # Widgets to be excluded from the grayed out area:
        self.__excluded_widgets: list[QW.QWidget] | None = None
        # Path to be grayed out:
        self.__path: QG.QPainterPath = QG.QPainterPath()

    def set_opacity_factor(self, opacity_factor: float) -> None:
        """
        Set the opacity factor of the grayed out area.

        Args:
            opacity_factor: Opacity factor.
        """
        self.__opacity_factor = opacity_factor

    def exclude(self, widgets: list[QW.QWidget]) -> None:
        """
        Exclude widgets from the grayed out area.

        Args:
            widgets: Widgets to be excluded.
        """
        self.__excluded_widgets = widgets

    def update_geometry(self) -> None:
        """
        Update the geometry of the widget.
        """
        if os.name == "nt":
            QW.QApplication.processEvents()
        self.setGeometry(self.parent().geometry())
        self.__path = QG.QPainterPath()
        # Path is defined as the rectangle of the main window minus the rectangles of
        # the excluded widgets:
        self.__path.addRect(QC.QRectF(self.parent().rect()))
        if self.__excluded_widgets is not None:
            for widget in self.__excluded_widgets:
                widget.raise_()
                widget.show()
                geometry = widget.frameGeometry()
                width, height = geometry.width(), geometry.height()
                point = widget.mapTo(self.parent(), QC.QPoint(0, 0))
                x, y = point.x(), point.y()
                widget_path = QG.QPainterPath()
                widget_path.addRect(QC.QRectF(x, y, width, height))
                self.__path -= QG.QPainterPath(widget_path)
        self.repaint()

    def showEvent(self, a0: QG.QShowEvent | None) -> None:
        """
        Event handler for the "show" event.

        Args:
            a0: Show event.
        """
        super().showEvent(a0)
        self.setGeometry(self.parent().geometry())

    def paintEvent(self, event: QG.QPaintEvent) -> None:
        """
        Event handler for the "paint" event.

        Args:
            event: Paint event.
        """
        super().paintEvent(event)
        painter = QG.QPainter(self)
        painter.setOpacity((0.75 if is_dark_theme() else 0.5) * self.__opacity_factor)
        painter.setBrush(QG.QBrush(QC.Qt.black))
        painter.setPen(QC.Qt.NoPen)
        painter.drawPath(self.__path)


class StepResult(enum.Enum):
    """
    Result of a step.
    """

    PREVIOUS = enum.auto()
    NEXT = enum.auto()
    CLOSE = enum.auto()
    DEMO = enum.auto()
    RESTART = enum.auto()


class StepDialog(QW.QDialog):
    """
    Base class for tour dialog boxes.

    Args:
        parent: Parent widget.
        step: Tour step.
    """

    def __init__(
        self,
        parent: QW.QWidget,
        step: TourStep,
    ) -> None:
        super().__init__(parent)
        self.setWindowTitle(step.title)
        self.setWindowFlags(QC.Qt.Popup)
        self.setModal(True)
        self.step = step
        self.result: StepResult | None = None
        self.__btnmap: dict[QW.QPushButton, StepResult] = {}
        self.__btnlayout = QW.QHBoxLayout()
        self._setup_ui()

    def __btn(self, text: str, result: StepResult, default: bool = False) -> None:
        """
        Create a button.

        Args:
            text: Button text.
            result: Result of the button.
            default: Default button. Defaults to False.
        """
        btn = QW.QPushButton(text)
        btn.setAutoDefault(default)
        btn.setDefault(default)
        icon = {
            "<<": "rewind.svg",
            "<": "previous.svg",
            ">": "next.svg",
            "X": "stop.svg",
        }.get(text)
        if icon is not None:
            btn.setIcon(get_icon(icon))
            btn.setText("")
            btn.setMinimumWidth(100)
        btn.clicked.connect(lambda: self.button_clicked(btn))
        self.__btnmap[btn] = result
        self.__btnlayout.addWidget(btn)

    def __btn_spacing(self, spacing: int) -> None:
        """
        Add spacing between buttons.

        Args:
            spacing: Spacing.
        """
        self.__btnlayout.addSpacing(spacing)

    def __create_label(
        self,
        text: str,
        delta_fontsize: int,
        color: QC.Qt.GlobalColor,
        bold: bool = False,
    ) -> QW.QLabel:
        """
        Create a label.

        Args:
            text: Label text.
            delta_fontsize: Delta font size.
            color: Text color.
            bold: Bold text.

        Returns:
            Label.
        """
        label = QW.QLabel(text)
        label.setWordWrap(True)
        label.setAlignment(QC.Qt.AlignCenter)
        # Change font size for something bigger:
        font = label.font()
        font.setPointSize(font.pointSize() + delta_fontsize)
        label.setFont(font)
        # Set text color:
        palette = label.palette()
        palette.setColor(QG.QPalette.WindowText, color)
        label.setPalette(palette)
        # Set bold text:
        font = label.font()
        font.setBold(bold)
        return label

    def __create_horizontal_line(self) -> QW.QFrame:
        """
        Create a horizontal line.

        Returns:
            Horizontal line.
        """
        line = QW.QFrame()
        line.setFrameShape(QW.QFrame.HLine)
        line.setFrameShadow(QW.QFrame.Sunken)
        line.setLineWidth(1)
        line.setMidLineWidth(0)
        palette = line.palette()
        palette.setColor(QG.QPalette.WindowText, QC.Qt.lightGray)
        line.setPalette(palette)
        return line

    def _setup_ui(self) -> None:
        """
        Setup the user interface.
        """
        if self.step.step_type == "introduction":
            self.__btn_spacing(50)
            self.__btn(_("Start"), StepResult.NEXT, True)
            self.__btn(_("Close"), StepResult.CLOSE)
            self.__btn_spacing(50)
        elif self.step.step_type == "conclusion":
            self.__btn("<<", StepResult.RESTART)
            self.__btn("<", StepResult.PREVIOUS)
            self.__btn(_("Demo"), StepResult.DEMO)
            self.__btn("X", StepResult.CLOSE, True)
        else:
            self.__btn("<<", StepResult.RESTART)
            self.__btn("<", StepResult.PREVIOUS)
            self.__btn(">", StepResult.NEXT, True)
            self.__btn("X", StepResult.CLOSE)
        self._layout = QW.QVBoxLayout()
        if self.step.step_type in ("introduction", "conclusion"):
            logo = QW.QLabel()
            logo.setPixmap(QG.QPixmap(get_image_file_path("DataLab-Banner2-100.png")))
            logo.setAlignment(QC.Qt.AlignCenter)
            self._layout.addWidget(logo)
        title_color = QC.Qt.lightGray if is_dark_theme() else QC.Qt.darkGray
        title = self.__create_label(self.step.title, 1, title_color, True)
        self._layout.addWidget(title)
        self._layout.addWidget(self.__create_horizontal_line())
        label_color = QC.Qt.cyan if is_dark_theme() else QC.Qt.darkBlue
        label_dsize = 3 if self.step.step_type == "regular" else 4
        label = self.__create_label(self.step.text, label_dsize, label_color)
        if self.step.step_type == "regular":
            label.setAlignment(QC.Qt.AlignLeft)
        self._layout.addWidget(label)
        self._layout.addSpacing(5)
        self._layout.addLayout(self.__btnlayout)
        if self.step.step_type == "introduction":
            self._layout.addSpacing(5)
            help_text = _(
                "Hit <b>Enter</b> to continue to the next step, or "
                "<b>Esc</b> to close the tour."
            )
            help_label = self.__create_label(help_text, -1, title_color)
            help_label.setAlignment(QC.Qt.AlignLeft)
            self._layout.addWidget(help_label)
        self.setLayout(self._layout)

    def button_clicked(self, button: QW.QPushButton) -> None:
        """
        Event handler for the "clicked" event on the buttons.

        Args:
            button: Button that was clicked.
        """
        self.result = self.__btnmap[button]
        if self.result is StepResult.CLOSE:
            self.reject()
        else:
            self.accept()

    def reject(self) -> None:
        """
        Event handler for the "reject" event.
        """
        self.result = StepResult.CLOSE
        super().reject()

    def paintEvent(self, event: QG.QPaintEvent) -> None:
        """
        Event handler for the "paint" event.

        Args:
            event: Paint event.
        """
        super().paintEvent(event)
        painter = QG.QPainter(self)
        painter.setOpacity(0.5)
        painter.setBrush(QG.QBrush(QC.Qt.black))
        painter.setPen(QC.Qt.NoPen)


class TourStep:
    """
    Tour step.

    Args:
        tour: Tour.
        title: Step title.
        text: Step text.
        widgets: Widgets to be highlighted. If None, no widget is highlighted, meaning
         that the step is probably an introduction or a conclusion.
        setup_callback: Callback function to be called before the step is displayed,
          which takes a single argument, the `CDLMainWindow` instance.
        teardown_callback: Callback function to be called after the step is displayed,
          which takes a single argument, the `CDLMainWindow` instance.
        step_type: Step type. Can be "regular", "introduction" or "conclusion".
    """

    def __init__(
        self,
        tour: BaseTour,
        title: str,
        text: str,
        widgets: list[QW.QWidget] | None = None,
        setup_callback: Callable | None = None,
        teardown_callback: Callable | None = None,
        step_type: str = "regular",
    ) -> None:
        self.tour = tour
        self.title = title
        self.text = text
        self.widgets = widgets
        self.setup_callback = setup_callback
        self.teardown_callback = teardown_callback
        self.step_type = step_type
        assert step_type in ["regular", "introduction", "conclusion"]

    def show(self) -> StepResult:
        """
        Show the step.

        Returns:
            Result of the step.
        """
        if self.setup_callback is not None:
            self.setup_callback(self.tour.win)
        self.update_cover(self.tour.cover)
        dialog = StepDialog(self.tour.win, self)
        if not execenv.unattended:
            dialog.exec()
        elif self.step_type != "conclusion":
            dialog.result = StepResult.NEXT
            dialog.accept()
        else:
            dialog.result = StepResult.CLOSE
            dialog.reject()
        if self.teardown_callback is not None:
            self.teardown_callback(self.tour.win)
        return dialog.result

    def update_cover(self, cover: Cover) -> None:
        """
        Update the cover widget.

        Args:
            cover: Cover widget.
        """
        cover.set_opacity_factor(1.0 if self.step_type == "regular" else 0.7)
        cover.exclude(self.widgets)
        cover.update_geometry()


class BaseTourMeta(type(QW.QWidget), abc.ABCMeta):
    """Mixed metaclass to avoid conflicts"""


class BaseTour(QW.QWidget, metaclass=BaseTourMeta):
    """
    Base class for the tour of DataLab features.

    Args:
        win: DataLab main window.
    """

    def __init__(self, win: CDLMainWindow) -> None:
        super().__init__(win)
        self.win = win
        self._steps: list[TourStep] = []
        self._current_step = 0
        self.cover = Cover(win)
        self.cover.show()
        self.__window_geometry: tuple[tuple[int, int], tuple[int, int]] | None = None
        self.__window_prepared = False
        self.setup_tour(win)

    def __resize_window(self, factor: float) -> None:
        """
        Resize the window so that it is centered on the screen and its size is
        `factor` times the size of the screen.

        Args:
            factor: Factor by which the size of the window is multiplied.
        """
        desktop = QW.QApplication.desktop()
        screen = desktop.screenNumber(desktop.cursor().pos())
        screen_geometry = desktop.screenGeometry(screen)
        width = int(screen_geometry.width() * factor)
        height = int(screen_geometry.height() * factor)
        self.win.resize(width, height)
        self.win.move(screen_geometry.center() - QC.QPoint(width // 2, height // 2))

    def __save_window_geometry(self) -> None:
        """Save the window geometry (size and position)."""
        self.__window_geometry = self.win.saveGeometry()

    def __restore_window_geometry(self) -> None:
        """Restore the window geometry (size and position)."""
        if self.__window_geometry is not None:
            self.win.restoreGeometry(self.__window_geometry)

    @abc.abstractmethod
    def setup_tour(self, win: CDLMainWindow) -> None:
        """
        Setup the tour: add steps to the tour.
        """

    @abc.abstractmethod
    def cleanup_tour(self, win: CDLMainWindow) -> None:
        """
        Cleanup the tour.
        """

    def add_step(
        self,
        title: str,
        text: str,
        widgets: list[QW.QWidget] | None = None,
        setup_callback: Callable | None = None,
        teardown_callback: Callable | None = None,
        step_type: str | None = None,
    ) -> None:
        """
        Add a step to the tour.

        Args:
            title: Step title.
            text: Step text.
            widgets: Widgets to be highlighted. If None, no widget is highlighted,
             meaning that the step is probably an introduction or a conclusion.
            setup_callback: Callback function to be called before the step is displayed,
             which takes a single argument, the `CDLMainWindow` instance.
            teardown_callback: Callback function to be called after the step is
             displayed, which takes a single argument, the `CDLMainWindow` instance.
            step_type: Step type. Can be "regular", "introduction" or "conclusion".
             Defaults to None.
        """
        if step_type is None:
            if self._steps:
                step_type = "regular"
            else:
                step_type = "introduction"
        step = TourStep(
            self, title, text, widgets, setup_callback, teardown_callback, step_type
        )
        self._steps.append(step)

    def show_current_step(self) -> None:
        """
        Show the current step.
        """
        step = self._steps[self._current_step]
        result = step.show()
        if result is StepResult.PREVIOUS:
            self.previous_step()
        elif result is StepResult.NEXT:
            self.next_step()
        elif result is StepResult.CLOSE:
            self.end()
        elif result is StepResult.DEMO:
            self.end()
            # pylint: disable=import-outside-toplevel
            from cdl.tests.scenarios import demo

            demo.play_demo(self.win)
        elif result is StepResult.RESTART:
            self.start()
        else:
            raise RuntimeError(f"Unknown result: {result}")

    def start(self) -> None:
        """
        Start the tour.
        """
        self._current_step = 0
        self.show_current_step()

    def end(self) -> None:
        """
        End the tour.
        """
        self.cover.close()
        self.cleanup_tour(self.win)
        self.__restore_window_geometry()

    def next_step(self) -> None:
        """
        Go to the next step.
        """
        if self._current_step == 0 and not self.__window_prepared:
            self.__save_window_geometry()
            self.__resize_window(0.7)
            self.cover.update_geometry()
            self.__window_prepared = True
        self._current_step += 1
        self.show_current_step()

    def previous_step(self) -> None:
        """
        Go to the previous step.
        """
        self._current_step -= 1
        self.show_current_step()


class Tour(BaseTour):
    """
    Tour of DataLab features.

    Args:
        win: DataLab main window.
    """

    def prepare_signalpanel(self, win: CDLMainWindow) -> None:
        """
        Prepare the signal panel.

        Args:
            win: DataLab main window.
        """
        # Create a signal:
        x = np.linspace(0, 10, 100)
        y = np.sin(x)
        sig = cdl.obj.create_signal(_("Test signal, for the tour"), x, y)
        win.add_object(sig)
        win.set_current_panel("signal")

    def prepare_imagepanel(self, win: CDLMainWindow) -> None:
        """
        Prepare the image panel.

        Args:
            win: DataLab main window.
        """
        # Create an image:
        x = np.linspace(0, 10, 100)
        y = np.linspace(0, 10, 100)
        xx, yy = np.meshgrid(x, y)
        zz = np.sin(xx) * np.cos(yy)
        img = cdl.obj.create_image(_("Test image, for the tour"), zz)
        win.add_object(img)
        win.set_current_panel("image")

    def popup_menu(self, win: CDLMainWindow, menu: QW.QMenu) -> None:
        """
        Popup a menu.

        Args:
            win: DataLab main window.
            menu_name: Name of the menu to popup.
        """
        menu.popup(win.mapToGlobal(QC.QPoint(50, 50)))

    def cleanup_tour(self, win: CDLMainWindow) -> None:
        """
        Cleanup the tour.

        Args:
            win: DataLab main window.
        """
        if len(win.signalpanel.objmodel) == 1:
            win.signalpanel.remove_all_objects()
        else:
            win.signalpanel.remove_object()
        if len(win.imagepanel.objmodel) == 1:
            win.imagepanel.remove_all_objects()
        else:
            win.imagepanel.remove_object()

    def setup_tour(self, win: CDLMainWindow) -> None:
        """
        Setup the tour: add steps to the tour.
        """
        self.add_step(
            _("Welcome to DataLab!"),
            _("This tour will guide you through the main features of DataLab."),
        )
        self.add_step(
            _("DataLab main window"),
            _(
                "This is the main window of DataLab. It is composed of several parts "
                "that we will describe in the following steps."
            ),
        )
        self.add_step(
            _("DataLab main window"),
            _(
                "Menus and toolbars regroup the main actions that can be performed "
                "in DataLab. Their content is adapted to the current panel, as we "
                "will see in the following steps."
            ),
            [
                win.menuBar(),
                win.main_toolbar,
                win.signalpanel_toolbar,
                win.imagepanel_toolbar,
                win.signalpanel.acthandler.view_toolbar,
                win.imagepanel.acthandler.view_toolbar,
            ],
        )
        self.add_step(
            _("DataLab main window"),
            _(
                "The main window is composed of two main panels: the Signal Panel, "
                "and the Image Panel.<br>"
                "Switching between panels is done using the highlighted tabs."
            ),
            [win.tabwidget.tabBar()],
        )
        self.add_step(
            _("Signal Panel"),
            _(
                "The <b>Signal Panel</b> is used to manage 1D signals."
                "It is composed of the elements shown in the following steps."
            ),
        )
        self.add_step(
            _("Signal Panel") + " – " + _("List and properties"),
            _(
                "In the highlighted area, signals are listed at the top, and their "
                "properties may be displayed and edited at the bottom.<br><br>"
                "Signals are numbered (but may be reorganized) and put together in "
                "numbered groups."
            ),
            [win.tabwidget.tabBar(), win.signalpanel],
            self.prepare_signalpanel,
        )
        self.add_step(
            _("Signal Panel") + " – " + _("View"),
            _(
                "Signals are plotted in the <b>Signal View</b>.<br><br>"
                "Curves may be customized using context menus or the vertical "
                "toolbar on the left (appearance settings are saved in the "
                "signal metadata)."
            ),
            [win.docks[win.signalpanel]],
        )
        self.add_step(
            _("Signal Panel") + " – " + _("File menu"),
            _(
                "The <b>File</b> menu contains actions to create new signals, "
                "import and export signals individually (various formats) or "
                "to save or restore the whole workspace (HDF5 files)."
            ),
            [win.menuBar()],
            lambda win: self.popup_menu(win, win.file_menu),
            lambda win: win.file_menu.hide(),
        )
        self.add_step(
            _("Signal Panel") + " – " + _("Edit menu"),
            _(
                "The <b>Edit</b> menu contains actions to edit signals individually "
                "or in groups."
            ),
            [win.menuBar()],
            lambda win: self.popup_menu(win, win.edit_menu),
            lambda win: win.edit_menu.hide(),
        )
        self.add_step(
            _("Signal Panel") + " – " + _("Operations menu"),
            _(
                "The <b>Operations</b> menu is focused on arithmetic operations, "
                "basic mathematical functions or data type conversions."
            ),
            [win.menuBar()],
            lambda win: self.popup_menu(win, win.operation_menu),
            lambda win: win.operation_menu.hide(),
        )
        self.add_step(
            _("Signal Panel") + " – " + _("Processing menu"),
            _("The <b>Processing</b> menu regroups 1->1 signal processing actions."),
            [win.menuBar()],
            lambda win: self.popup_menu(win, win.processing_menu),
            lambda win: win.processing_menu.hide(),
        )
        self.add_step(
            _("Signal Panel") + " – " + _("Analysis menu"),
            _(
                "The <b>Analysis</b> menu regroups 1->0 signal computing actions "
                "(that is, actions that do not modify the signals, but compute "
                "a result, e.g. scalar values), with optional ROI selection."
            ),
            [win.menuBar()],
            lambda win: self.popup_menu(win, win.analysis_menu),
            lambda win: win.analysis_menu.hide(),
        )
        self.add_step(
            _("Image Panel"),
            _(
                "The <b>Image Panel</b> is used to manage 2D images. It is composed "
                "of the elements shown in the following steps."
            ),
            [],
            lambda win: win.set_current_panel("image"),
        )
        self.add_step(
            _("Image Panel") + " – " + _("List and properties"),
            _(
                "In the highlighted area, images are listed at the top, and their "
                "properties may be displayed and edited at the bottom.<br><br>"
                "Images are numbered (but may be reorganized) and put together in "
                "numbered groups."
            ),
            [win.tabwidget.tabBar(), win.imagepanel],
            self.prepare_imagepanel,
        )
        self.add_step(
            _("Image Panel") + " – " + _("View"),
            _(
                "Images are shown in the <b>Image View</b>.<br><br>"
                "The displayed images may be customized using context menus "
                "or the vertical toolbar on the left (appearance settings "
                "are saved in the image metadata)."
            ),
            [win.docks[win.imagepanel]],
        )
        self.add_step(
            _("Image Panel") + " – " + _("File menu"),
            _(
                "The <b>File</b> menu contains actions to create new images, "
                "import and export images individually (various formats) or "
                "to save or restore the whole workspace (HDF5 files)."
            ),
            [win.menuBar()],
            lambda win: self.popup_menu(win, win.file_menu),
            lambda win: win.file_menu.hide(),
        )
        self.add_step(
            _("Image Panel") + " – " + _("Edit menu"),
            _(
                "The <b>Edit</b> menu contains actions to edit images individually "
                "or in groups."
            ),
            [win.menuBar()],
            lambda win: self.popup_menu(win, win.edit_menu),
            lambda win: win.edit_menu.hide(),
        )
        self.add_step(
            _("Image Panel") + " – " + _("Operations menu"),
            _(
                "The <b>Operations</b> menu is focused on arithmetic operations, "
                "data type conversions, pixel binning resize, intensity profiles, ..."
            ),
            [win.menuBar()],
            lambda win: self.popup_menu(win, win.operation_menu),
            lambda win: win.operation_menu.hide(),
        )
        self.add_step(
            _("Image Panel") + " – " + _("Processing menu"),
            _(
                "The <b>Processing</b> menu regroups 1->1 image processing actions "
                "(that is, actions that modify the images)."
            ),
            [win.menuBar()],
            lambda win: self.popup_menu(win, win.processing_menu),
            lambda win: win.processing_menu.hide(),
        )
        self.add_step(
            _("Image Panel") + " – " + _("Analysis menu"),
            _(
                "The <b>Analysis</b> menu regroups 1->0 image computing actions "
                "(that is, actions that do not modify the images, but compute "
                "a result, e.g. circle coordinates), with optional ROI selection."
            ),
            [win.menuBar()],
            lambda win: self.popup_menu(win, win.analysis_menu),
            lambda win: win.analysis_menu.hide(),
        )
        self.add_step(
            _("Extensions"),
            _(
                "DataLab is designed to be easily extended with new features, "
                "by using <b>Macros</b>, <b>Plugins</b> or <b>Remote Control</b>."
                "<br><br>The common point between these mechanisms is that "
                "they are all relying on Python scripts and the <i>DataLab API</i>."
            ),
        )
        self.add_step(
            _("Extensions") + " – " + _("Macros"),
            _(
                "The <b>Macro Panel</b> allows to create, edit and run macros.<br>"
                "Macros are saved together with the DataLab workspace (HDF5 file)."
            ),
            [win.docks[win.macropanel]],
            lambda win: win.set_current_panel("macro"),
            lambda win: win.set_current_panel("signal"),
        )
        self.add_step(
            _("Extensions") + " – " + _("Plugins"),
            _(
                "The <b>Plugins</b> menu regroups features that are not part of "
                "the core of DataLab, but that are provided as plugins.<br>"
                "(See the documentation for more information about plugins.)"
            ),
            [win.menuBar()],
            lambda win: self.popup_menu(win, win.plugins_menu),
            lambda win: win.plugins_menu.hide(),
        )
        self.add_step(
            _("This is the end of the tour!"),
            _("You can show the tour again, or close this dialog box."),
            step_type="conclusion",
        )


def start(win: CDLMainWindow) -> None:
    """
    Start the tour of DataLab features.

    Args:
        win: DataLab main window.
    """
    tour = Tour(win)
    tour.start()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Memory status widget application test
"""

# guitest: show

import numpy as np
import psutil

from cdl import config
from cdl.env import execenv
from cdl.obj import Gauss2DParam, ImageTypes, new_image_param
from cdl.tests import cdltest_app_context


def memory_alarm(threshold):
    """Memory alarm test"""
    config.Conf.main.available_memory_threshold.set(threshold)
    rng = np.random.default_rng()
    with cdltest_app_context() as win:
        panel = win.imagepanel
        win.memorystatus.update_status()  # Force memory status update
        newparam = new_image_param(itype=ImageTypes.GAUSS)
        addparam = Gauss2DParam.create(
            x0=rng.integers(-9, 9), y0=rng.integers(-9, 9), sigma=rng.integers(1, 20)
        )
        panel.new_object(newparam, addparam=addparam, edit=False)


def test_mem_status():
    """Memory alarm test"""
    mem_available = psutil.virtual_memory().available // (1024**2)
    execenv.print(f"Memory status widget test (memory available: {mem_available} MB):")
    for index, threshold in enumerate((mem_available * 2, mem_available - 100)):
        execenv.print(f"    Threshold {index}: {threshold} MB")
        memory_alarm(threshold)
    config.reset()


if __name__ == "__main__":
    test_mem_status()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
.. Image panel (see parent package :mod:`cdl.core.gui.panel`)
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

from typing import TYPE_CHECKING, Type

from plotpy.tools import (
    AnnotatedCircleTool,
    AnnotatedEllipseTool,
    AnnotatedPointTool,
    AnnotatedRectangleTool,
    AnnotatedSegmentTool,
    LabelTool,
)

from cdl.config import Conf, _
from cdl.core.gui import roieditor
from cdl.core.gui.actionhandler import ImageActionHandler
from cdl.core.gui.panel.base import BaseDataPanel
from cdl.core.gui.plothandler import ImagePlotHandler
from cdl.core.gui.processor.image import ImageProcessor
from cdl.core.io.image import ImageIORegistry
from cdl.core.model.image import (
    ImageDatatypes,
    ImageObj,
    ImageROI,
    create_image_from_param,
    new_image_param,
)

if TYPE_CHECKING:
    import guidata.dataset as gds
    from plotpy.plot import BasePlot
    from qtpy import QtWidgets as QW

    from cdl.core.gui.docks import DockablePlotWidget
    from cdl.core.model.image import NewImageParam


class ImagePanel(BaseDataPanel[ImageObj, ImageROI, roieditor.ImageROIEditor]):
    """Object handling the item list, the selected item properties and plot,
    specialized for Image objects"""

    PANEL_STR = _("Image Panel")
    PANEL_STR_ID = "image"
    PARAMCLASS = ImageObj
    MINDIALOGSIZE = (800, 800)

    # The following tools are used to create annotations on images. The annotation
    # items are created using PlotPy's default settings. Those appearance settings
    # may be modified in the configuration (see `cdl.config`).
    ANNOTATION_TOOLS = (
        AnnotatedCircleTool,
        AnnotatedSegmentTool,
        AnnotatedRectangleTool,
        AnnotatedPointTool,
        AnnotatedEllipseTool,
        LabelTool,
    )

    IO_REGISTRY = ImageIORegistry
    H5_PREFIX = "DataLab_Ima"
    ROIDIALOGOPTIONS = {"show_itemlist": True, "show_contrast": False}

    # pylint: disable=duplicate-code

    @staticmethod
    def get_roieditor_class() -> Type[roieditor.ImageROIEditor]:
        """Return ROI editor class"""
        return roieditor.ImageROIEditor

    def __init__(
        self,
        parent: QW.QWidget,
        dockableplotwidget: DockablePlotWidget,
        panel_toolbar: QW.QToolBar,
    ) -> None:
        super().__init__(parent)
        self.plothandler = ImagePlotHandler(self, dockableplotwidget.plotwidget)
        self.processor = ImageProcessor(self, dockableplotwidget.plotwidget)
        view_toolbar = dockableplotwidget.toolbar
        self.acthandler = ImageActionHandler(self, panel_toolbar, view_toolbar)

    # ------Refreshing GUI--------------------------------------------------------------
    def plot_lut_changed(self, plot: BasePlot) -> None:
        """The LUT of the plot has changed: updating image objects accordingly

        Args:
            plot: Plot object
        """
        zmin, zmax = plot.get_axis_limits(plot.colormap_axis)
        for obj in self.objview.get_sel_objects():
            obj.zscalemin, obj.zscalemax = zmin, zmax
            if obj is self.objview.get_current_object():
                self.objprop.update_properties_from(obj)

    # ------Creating, adding, removing objects------------------------------------------
    def get_newparam_from_current(
        self, newparam: NewImageParam | None = None, title: str | None = None
    ) -> NewImageParam | None:
        """Get new object parameters from the current object.

        Args:
            newparam (guidata.dataset.DataSet): new object parameters.
             If None, create a new one.
            title: new object title. If None, use the current object title, or the
             default title.

        Returns:
            New object parameters
        """
        curobj: ImageObj = self.objview.get_current_object()
        if newparam is None:
            newparam = new_image_param(title=title)
        if curobj is not None:
            newparam.width, newparam.height = curobj.data.shape
            newparam.dtype = ImageDatatypes.from_dtype(curobj.data.dtype)
        return newparam

    def new_object(
        self,
        newparam: NewImageParam | None = None,
        addparam: gds.DataSet | None = None,
        edit: bool = True,
        add_to_panel: bool = True,
    ) -> ImageObj | None:
        """Create a new object (image).

        Args:
            newparam (Daguidata.dataset.datatypes.DataSettaSet): new object parameters
            addparam (guidata.dataset.DataSet): additional parameters
            edit (bool): Open a dialog box to edit parameters (default: True)
            add_to_panel (bool): Add the object to the panel (default: True)

        Returns:
            New object
        """
        if not self.mainwindow.confirm_memory_state():
            return None
        newparam = self.get_newparam_from_current(newparam)
        image = create_image_from_param(
            newparam, addparam=addparam, edit=edit, parent=self.parent()
        )
        if image is None:
            return None
        if add_to_panel:
            self.add_object(image)
        return image

    def toggle_show_contrast(self, state: bool) -> None:
        """Toggle show contrast option"""
        Conf.view.show_contrast.set(state)
        self.SIG_REFRESH_PLOT.emit("selected", True)



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
.. Common computation objects (see parent package :mod:`cdl.computation`)
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

# Note:
# ----
# All dataset classes must also be imported in the cdl.computation.param module.

from __future__ import annotations

from typing import TYPE_CHECKING

import guidata.dataset as gds
import numpy as np

from cdl.config import _
from cdl.obj import ResultProperties, create_signal

if TYPE_CHECKING:
    from typing import Callable

    from cdl.obj import ImageObj, SignalObj


class ArithmeticParam(gds.DataSet):
    """Arithmetic parameters"""

    def get_operation(self) -> str:
        """Return the operation string"""
        o, a, b = self.operator, self.factor, self.constant
        b_added = False
        if a == 0.0:
            if o in ("+", "-"):
                txt = "obj3 = obj1"
            elif b == 0.0:
                txt = "obj3 = 0"
            else:
                txt = f"obj3 = {b}"
                b_added = True
        elif a == 1.0:
            txt = f"obj3 = obj1 {o} obj2"
        else:
            txt = f"obj3 = (obj1 {o} obj2) × {a}"
        if b != 0.0 and not b_added:
            txt += f" + {b}"
        return txt

    def update_operation(self, _item, _value):  # pylint: disable=unused-argument
        """Update the operation item"""
        self.operation = self.get_operation()

    operators = ("+", "-", "×", "/")
    operator = gds.ChoiceItem(_("Operator"), list(zip(operators, operators))).set_prop(
        "display", callback=update_operation
    )
    factor = (
        gds.FloatItem(_("Factor"), default=1.0)
        .set_pos(col=1)
        .set_prop("display", callback=update_operation)
    )
    constant = (
        gds.FloatItem(_("Constant"), default=0.0)
        .set_pos(col=1)
        .set_prop("display", callback=update_operation)
    )
    operation = gds.StringItem(_("Operation"), default="").set_prop(
        "display", active=False
    )
    restore_dtype = gds.BoolItem(
        _("Convert to `obj1` data type"), label=_("Result"), default=True
    )


class GaussianParam(gds.DataSet):
    """Gaussian filter parameters"""

    sigma = gds.FloatItem("σ", default=1.0)


HELP_MODE = _("""Mode of the filter:
- 'reflect': Reflect the data at the boundary
- 'constant': Pad with a constant value
- 'nearest': Pad with the nearest value
- 'mirror': Reflect the data at the boundary with the data itself
- 'wrap': Circular boundary""")


class MovingAverageParam(gds.DataSet):
    """Moving average parameters"""

    n = gds.IntItem(_("Size of the moving window"), default=3, min=1)
    modes = ("reflect", "constant", "nearest", "mirror", "wrap")
    mode = gds.ChoiceItem(
        _("Mode"), list(zip(modes, modes)), default="reflect", help=HELP_MODE
    )


class MovingMedianParam(gds.DataSet):
    """Moving median parameters"""

    n = gds.IntItem(_("Size of the moving window"), default=3, min=1, even=False)
    modes = ("reflect", "constant", "nearest", "mirror", "wrap")
    mode = gds.ChoiceItem(
        _("Mode"), list(zip(modes, modes)), default="nearest", help=HELP_MODE
    )


class ClipParam(gds.DataSet):
    """Data clipping parameters"""

    lower = gds.FloatItem(_("Lower clipping value"), check=False)
    upper = gds.FloatItem(_("Upper clipping value"), check=False)


class NormalizeParam(gds.DataSet):
    """Normalize parameters"""

    methods = (
        ("maximum", _("Maximum")),
        ("amplitude", _("Amplitude")),
        ("area", _("Area")),
        ("energy", _("Energy")),
        ("rms", _("RMS")),
    )
    method = gds.ChoiceItem(_("Normalize with respect to"), methods)


class HistogramParam(gds.DataSet):
    """Histogram parameters"""

    def get_suffix(self, data: np.ndarray) -> str:
        """Return suffix for the histogram computation

        Args:
            data: data array
        """
        suffix = f"bins={self.bins:d}"
        if self.lower is not None:
            suffix += f", ymin={self.lower:.3f}"
        else:
            self.lower = np.min(data)
        if self.upper is not None:
            suffix += f", ymax={self.upper:.3f}"
        else:
            self.upper = np.max(data)

    bins = gds.IntItem(_("Number of bins"), default=256, min=1)
    lower = gds.FloatItem(_("Lower limit"), default=None, check=False)
    upper = gds.FloatItem(_("Upper limit"), default=None, check=False)


class FFTParam(gds.DataSet):
    """FFT parameters"""

    shift = gds.BoolItem(_("Shift"), help=_("Shift zero frequency to center"))


class SpectrumParam(gds.DataSet):
    """Spectrum parameters"""

    log = gds.BoolItem(_("Logarithmic scale"), default=False)


class ConstantParam(gds.DataSet):
    """Parameter used to set a constant value to used in operations"""

    value = gds.FloatItem(_("Constant value"))


# MARK: Helper functions for creating result objects -----------------------------------


def dst_11(
    src: SignalObj | ImageObj, name: str, suffix: str | None = None
) -> SignalObj | ImageObj:
    """Create a result object, as returned by the callback function of the
    :func:`cdl.core.gui.processor.base.BaseProcessor.compute_11` method

    Args:
        src: source signal or image object
        name: name of the function. If provided, the title of the result object
         will be `{name}({src.short_id})|{suffix})`, unless the name is a single
         character, in which case the title will be `{src.short_id}{name}{suffix}`
         where `name` is an operator and `suffix` is the other term of the operation.
        suffix: suffix to add to the title. Optional.

    Returns:
        Result signal or image object
    """
    if len(name) == 1:  # This is an operator
        title = f"{src.short_id}{name}"
    else:
        title = f"{name}({src.short_id})"
        if suffix:  # suffix may be None or an empty string
            title += "|"
    if suffix:  # suffix may be None or an empty string
        title += suffix
    return src.copy(title=title)


def dst_n1n(
    src1: SignalObj | ImageObj,
    src2: SignalObj | ImageObj,
    name: str,
    suffix: str | None = None,
) -> SignalObj | ImageObj:
    """Create a result  object, as returned by the callback function of the
    :func:`cdl.core.gui.processor.base.BaseProcessor.compute_n1n` method

    Args:
        src1: input signal or image object
        src2: input signal or image object
        name: name of the processing function

    Returns:
        Output signal or image object
    """
    if len(name) == 1:  # This is an operator
        title = f"{src1.short_id}{name}{src2.short_id}"
    else:
        title = f"{name}({src1.short_id}, {src2.short_id})"
    if suffix is not None:
        title += "|" + suffix
    return src1.copy(title=title)


def new_signal_result(
    src: SignalObj | ImageObj,
    name: str,
    suffix: str | None = None,
    units: tuple[str, str] | None = None,
    labels: tuple[str, str] | None = None,
) -> SignalObj:
    """Create new signal object as a result of a compute_11 function

    As opposed to the `dst_11` functions, this function creates a new signal object
    without copying the original object metadata, except for the "source" entry.

    Args:
        src: input signal or image object
        name: name of the processing function
        suffix: suffix to add to the title
        units: units of the output signal
        labels: labels of the output signal

    Returns:
        Output signal object
    """
    title = f"{name}({src.short_id})"
    dst = create_signal(title=title, units=units, labels=labels)
    if suffix is not None:
        dst.title += "|" + suffix
    if "source" in src.metadata:
        dst.metadata["source"] = src.metadata["source"]  # Keep track of the source
    return dst


def calc_resultproperties(
    title: str, obj: SignalObj | ImageObj, labeledfuncs: dict[str, Callable]
) -> ResultProperties:
    """Calculate result properties by executing a computation function
    on a signal/image object.

    Args:
        title: title of the result properties
        obj: signal or image object
        labeledfuncs: dictionary of labeled computation functions. The keys are
         the labels of the computation functions and the values are the functions
         themselves (each function must take a single argument - which is the data
         of the ROI or the whole signal/image - and return a float)

    Returns:
        Result properties object
    """
    if not all(isinstance(k, str) for k in labeledfuncs.keys()):
        raise ValueError("Keys of labeledfuncs must be strings")
    if not all(callable(v) for v in labeledfuncs.values()):
        raise ValueError("Values of labeledfuncs must be functions")

    res = []
    for i_roi in [None] + list(obj.iterate_roi_indices()):
        data_roi = obj.get_data(i_roi)
        val_roi = -1 if i_roi is None else i_roi
        res.append([val_roi] + [fn(data_roi) for fn in labeledfuncs.values()])
    return ResultProperties(title, np.array(res), list(labeledfuncs.keys()))



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
.. Image Processing Algorithms (see parent package :mod:`cdl.algorithms`)
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

from typing import Literal

import numpy as np
import scipy.ndimage as spi
import scipy.spatial as spt
from numpy import ma
from skimage import exposure, feature, measure, transform

# MARK: Level adjustment ---------------------------------------------------------------


def scale_data_to_min_max(
    data: np.ndarray, zmin: float | int, zmax: float | int
) -> np.ndarray:
    """Scale array `data` to fit [zmin, zmax] dynamic range

    Args:
        data: Input data
        zmin: Minimum value of output data
        zmax: Maximum value of output data

    Returns:
        Scaled data
    """
    dmin = data.min()
    dmax = data.max()
    fdata = np.array(data, dtype=float)
    fdata -= dmin
    fdata *= float(zmax - zmin) / (dmax - dmin)
    fdata += float(zmin)
    return np.array(fdata, data.dtype)


def normalize(
    data: np.ndarray,
    parameter: Literal["maximum", "amplitude", "area", "energy", "rms"] = "maximum",
) -> np.ndarray:
    """Normalize input array to a given parameter.

    Args:
        data: Input data
        parameter: Normalization parameter (default: "maximum")

    Returns:
        Normalized array
    """
    if parameter == "maximum":
        return scale_data_to_min_max(data, data.min() / data.max(), 1.0)
    if parameter == "amplitude":
        return scale_data_to_min_max(data, 0.0, 1.0)
    fdata = np.array(data, dtype=float)
    if parameter == "area":
        return fdata / fdata.sum()
    if parameter == "energy":
        return fdata / np.sqrt(np.sum(fdata * fdata.conjugate()))
    if parameter == "rms":
        return fdata / np.sqrt(np.mean(fdata * fdata.conjugate()))
    raise ValueError(f"Unsupported parameter {parameter}")


# MARK: Fourier analysis ---------------------------------------------------------------


def fft2d(z: np.ndarray, shift: bool = True) -> np.ndarray:
    """Compute FFT of complex array `z`

    Args:
        z: Input data
        shift: Shift zero frequency to center (default: True)

    Returns:
        FFT of input data
    """
    z1 = np.fft.fft2(z)
    if shift:
        z1 = np.fft.fftshift(z1)
    return z1


def ifft2d(z: np.ndarray, shift: bool = True) -> np.ndarray:
    """Compute inverse FFT of complex array `z`

    Args:
        z: Input data
        shift: Shift zero frequency to center (default: True)

    Returns:
        Inverse FFT of input data
    """
    if shift:
        z = np.fft.ifftshift(z)
    z1 = np.fft.ifft2(z)
    return z1


def magnitude_spectrum(z: np.ndarray, log_scale: bool = False) -> np.ndarray:
    """Compute magnitude spectrum of complex array `z`

    Args:
        z: Input data
        log_scale: Use log scale (default: False)

    Returns:
        Magnitude spectrum of input data
    """
    z1 = np.abs(fft2d(z))
    if log_scale:
        z1 = 20 * np.log10(z1.clip(1e-10))
    return z1


def phase_spectrum(z: np.ndarray) -> np.ndarray:
    """Compute phase spectrum of complex array `z`

    Args:
        z: Input data

    Returns:
        Phase spectrum of input data (in degrees)
    """
    return np.rad2deg(np.angle(fft2d(z)))


def psd(z: np.ndarray, log_scale: bool = False) -> np.ndarray:
    """Compute power spectral density of complex array `z`

    Args:
        z: Input data
        log_scale: Use log scale (default: False)

    Returns:
        Power spectral density of input data
    """
    z1 = np.abs(fft2d(z)) ** 2
    if log_scale:
        z1 = 10 * np.log10(z1.clip(1e-10))
    return z1


# MARK: Binning ------------------------------------------------------------------------


BINNING_OPERATIONS = ("sum", "average", "median", "min", "max")


def binning(
    data: np.ndarray,
    sx: int,
    sy: int,
    operation: Literal["sum", "average", "median", "min", "max"],
    dtype=None,
) -> np.ndarray:
    """Perform image pixel binning

    Args:
        data: Input data
        sx: Binning size along x (number of pixels to bin together)
        sy: Binning size along y (number of pixels to bin together)
        operation: Binning operation
        dtype: Output data type (default: None, i.e. same as input)

    Returns:
        Binned data
    """
    ny, nx = data.shape
    shape = (ny // sy, sy, nx // sx, sx)
    try:
        bdata = data[: ny - ny % sy, : nx - nx % sx].reshape(shape)
    except ValueError as err:
        raise ValueError("Binning is not a multiple of image dimensions") from err
    if operation == "sum":
        bdata = np.array(bdata, dtype=float).sum(axis=(-1, 1))
    elif operation == "average":
        bdata = bdata.mean(axis=(-1, 1))
    elif operation == "median":
        bdata = ma.median(bdata, axis=(-1, 1))
    elif operation == "min":
        bdata = bdata.min(axis=(-1, 1))
    elif operation == "max":
        bdata = bdata.max(axis=(-1, 1))
    else:
        valid = ", ".join(BINNING_OPERATIONS)
        raise ValueError(f"Invalid operation {operation} (valid values: {valid})")
    return np.array(bdata, dtype=data.dtype if dtype is None else np.dtype(dtype))


# MARK: Background subtraction ---------------------------------------------------------


def flatfield(
    rawdata: np.ndarray, flatdata: np.ndarray, threshold: float | None = None
) -> np.ndarray:
    """Compute flat-field correction

    Args:
        rawdata: Raw data
        flatdata: Flat-field data
        threshold: Threshold for flat-field correction (default: None)

    Returns:
        Flat-field corrected data
    """
    dtemp = np.array(rawdata, dtype=float, copy=True) * flatdata.mean()
    dunif = np.array(flatdata, dtype=float, copy=True)
    dunif[dunif == 0] = 1.0
    dcorr_all = np.array(dtemp / dunif, dtype=rawdata.dtype)
    dcorr = np.array(rawdata, copy=True)
    dcorr[rawdata > threshold] = dcorr_all[rawdata > threshold]
    return dcorr


# MARK: Misc. analyses -----------------------------------------------------------------


def get_centroid_fourier(data: np.ndarray) -> tuple[float, float]:
    """Return image centroid using Fourier algorithm

    Args:
        data: Input data

    Returns:
        Centroid coordinates (row, col)
    """
    # Fourier transform method as discussed by Weisshaar et al.
    # (http://www.mnd-umwelttechnik.fh-wiesbaden.de/pig/weisshaar_u5.pdf)
    rows, cols = data.shape
    if rows == 1 or cols == 1:
        return 0, 0

    i = np.arange(0, rows).reshape(1, rows)
    sin_a = np.sin((i - 1) * 2 * np.pi / (rows - 1)).T
    cos_a = np.cos((i - 1) * 2 * np.pi / (rows - 1)).T

    j = np.arange(0, cols).reshape(cols, 1)
    sin_b = np.sin((j - 1) * 2 * np.pi / (cols - 1)).T
    cos_b = np.cos((j - 1) * 2 * np.pi / (cols - 1)).T

    a = (cos_a * data).sum()
    b = (sin_a * data).sum()
    c = (data * cos_b).sum()
    d = (data * sin_b).sum()

    rphi = (0 if b > 0 else 2 * np.pi) if a > 0 else np.pi
    cphi = (0 if d > 0 else 2 * np.pi) if c > 0 else np.pi

    if a * c == 0.0:
        return 0, 0

    row = (np.arctan(b / a) + rphi) * (rows - 1) / (2 * np.pi) + 1
    col = (np.arctan(d / c) + cphi) * (cols - 1) / (2 * np.pi) + 1

    row = np.nan if row is np.ma.masked else row
    col = np.nan if col is np.ma.masked else col

    return row, col


def get_absolute_level(data: np.ndarray, level: float) -> float:
    """Return absolute level

    Args:
        data: Input data
        level: Relative level (0.0 to 1.0)

    Returns:
        Absolute level
    """
    if not isinstance(level, float) or level < 0.0 or level > 1.0:
        raise ValueError("Level must be a float between 0. and 1.")
    return (float(np.nanmin(data)) + float(np.nanmax(data))) * level


def get_enclosing_circle(
    data: np.ndarray, level: float = 0.5
) -> tuple[int, int, float]:
    """Return (x, y, radius) for the circle contour enclosing image
    values above threshold relative level (.5 means FWHM)

    Args:
        data: Input data
        level: Relative level (default: 0.5)

    Returns:
        A tuple (x, y, radius)

    Raises:
        ValueError: No contour was found
    """
    data_th = data.copy()
    data_th[data <= get_absolute_level(data, level)] = 0.0
    contours = measure.find_contours(data_th)
    model = measure.CircleModel()
    result = None
    max_radius = 1.0
    for contour in contours:
        if model.estimate(contour):
            yc, xc, radius = model.params
            if radius > max_radius:
                result = (int(xc), int(yc), radius)
                max_radius = radius
    if result is None:
        raise ValueError("No contour was found")
    return result


def get_radial_profile(
    data: np.ndarray, center: tuple[int, int]
) -> tuple[np.ndarray, np.ndarray]:
    """Return radial profile of image data

    Args:
        data: Input data (2D array)
        center: Coordinates of the center of the profile (x, y)

    Returns:
        Radial profile (X, Y) where X is the distance from the center (1D array)
        and Y is the average value of pixels at this distance (1D array)
    """
    y, x = np.indices((data.shape))  # Get the indices of pixels
    x0, y0 = center
    r = np.sqrt((x - x0) ** 2 + (y - y0) ** 2)  # Calculate distance to the center
    r = r.astype(int)

    # Average over the same distance
    tbin = np.bincount(r.ravel(), data.ravel())  # Sum of pixel values at each distance
    nr = np.bincount(r.ravel())  # Number of pixels at each distance

    yprofile = tbin / nr  # this is the half radial profile
    # Let's mirror it to get the full radial profile (the first element is the center)
    yprofile = np.concatenate((yprofile[::-1], yprofile[1:]))
    # The x axis is the distance from the center (0 is the center)
    xprofile = np.arange(len(yprofile)) - len(yprofile) // 2

    return xprofile, yprofile


def distance_matrix(coords: list) -> np.ndarray:
    """Return distance matrix from coords

    Args:
        coords: List of coordinates

    Returns:
        Distance matrix
    """
    return np.triu(spt.distance.cdist(coords, coords, "euclidean"))


def get_2d_peaks_coords(
    data: np.ndarray, size: int | None = None, level: float = 0.5
) -> np.ndarray:
    """Detect peaks in image data, return coordinates.

    If neighborhoods size is None, default value is the highest value
    between 50 pixels and the 1/40th of the smallest image dimension.

    Detection threshold level is relative to difference
    between data maximum and minimum values.

    Args:
        data: Input data
        size: Neighborhood size (default: None)
        level: Relative level (default: 0.5)

    Returns:
        Coordinates of peaks
    """
    if size is None:
        size = max(min(data.shape) // 40, 50)
    data_max = spi.maximum_filter(data, size)
    data_min = spi.minimum_filter(data, size)
    data_diff = data_max - data_min
    diff = (data_max - data_min) > get_absolute_level(data_diff, level)
    maxima = data == data_max
    maxima[diff == 0] = 0
    labeled, _num_objects = spi.label(maxima)
    slices = spi.find_objects(labeled)
    coords = []
    for dy, dx in slices:
        x_center = int(0.5 * (dx.start + dx.stop - 1))
        y_center = int(0.5 * (dy.start + dy.stop - 1))
        coords.append((x_center, y_center))
    if len(coords) > 1:
        # Eventually removing duplicates
        dist = distance_matrix(coords)
        for index in reversed(np.unique(np.where((dist < size) & (dist > 0))[1])):
            coords.pop(index)
    return np.array(coords)


def get_contour_shapes(
    data: np.ndarray | ma.MaskedArray,
    shape: Literal["circle", "ellipse", "polygon"] = "ellipse",
    level: float = 0.5,
) -> np.ndarray:
    """Find iso-valued contours in a 2D array, above relative level (.5 means FWHM),
    then fit contours with shape ('ellipse' or 'circle')

    Args:
        data: Input data
        shape: Shape to fit. Default is 'ellipse'
        level: Relative level (default: 0.5)

    Returns:
        Coordinates of shapes
    """
    # pylint: disable=too-many-locals
    assert shape in ("circle", "ellipse", "polygon")
    contours = measure.find_contours(data, level=get_absolute_level(data, level))
    coords = []
    for contour in contours:
        # `contour` is a (N, 2) array (rows, cols): we need to check if all those
        # coordinates are masked: if so, we skip this contour
        if isinstance(data, ma.MaskedArray) and np.all(
            data.mask[contour[:, 0].astype(int), contour[:, 1].astype(int)]
        ):
            continue
        if shape == "circle":
            model = measure.CircleModel()
            if model.estimate(contour):
                yc, xc, r = model.params
                if r <= 1.0:
                    continue
                coords.append([xc, yc, r])
        elif shape == "ellipse":
            model = measure.EllipseModel()
            if model.estimate(contour):
                yc, xc, b, a, theta = model.params
                if a <= 1.0 or b <= 1.0:
                    continue
                coords.append([xc, yc, a, b, theta])
        elif shape == "polygon":
            # `contour` is a (N, 2) array (rows, cols): we need to convert it
            # to a list of x, y coordinates flattened in a single list
            coords.append(contour[:, ::-1].flatten())
        else:
            raise NotImplementedError(f"Invalid contour model {model}")
    if shape == "polygon":
        # `coords` is a list of arrays of shape (N, 2) where N is the number of points
        # that can vary from one array to another, so we need to padd with NaNs each
        # array to get a regular array:
        max_len = max(coord.shape[0] for coord in coords)
        arr = np.full((len(coords), max_len), np.nan)
        for i_row, coord in enumerate(coords):
            arr[i_row, : coord.shape[0]] = coord
        return arr
    return np.array(coords)


def get_hough_circle_peaks(
    data: np.ndarray,
    min_radius: float | None = None,
    max_radius: float | None = None,
    nb_radius: int | None = None,
    min_distance: int = 1,
) -> np.ndarray:
    """Detect peaks in image from circle Hough transform, return circle coordinates.

    Args:
        data: Input data
        min_radius: Minimum radius (default: None)
        max_radius: Maximum radius (default: None)
        nb_radius: Number of radii (default: None)
        min_distance: Minimum distance between circles (default: 1)

    Returns:
        Coordinates of circles
    """
    assert min_radius is not None and max_radius is not None and max_radius > min_radius
    if nb_radius is None:
        nb_radius = max_radius - min_radius + 1
    hough_radii = np.arange(
        min_radius, max_radius + 1, (max_radius - min_radius + 1) // nb_radius
    )
    hough_res = transform.hough_circle(data, hough_radii)
    _accums, cx, cy, radii = transform.hough_circle_peaks(
        hough_res, hough_radii, min_xdistance=min_distance, min_ydistance=min_distance
    )
    return np.vstack([cx, cy, radii]).T


# MARK: Blob detection -----------------------------------------------------------------


def __blobs_to_coords(blobs: np.ndarray) -> np.ndarray:
    """Convert blobs to coordinates

    Args:
        blobs: Blobs

    Returns:
        Coordinates
    """
    cy, cx, radii = blobs.T
    coords = np.vstack([cx, cy, radii]).T
    return coords


def find_blobs_dog(
    data: np.ndarray,
    min_sigma: float = 1,
    max_sigma: float = 30,
    overlap: float = 0.5,
    threshold_rel: float = 0.2,
    exclude_border: bool = True,
) -> np.ndarray:
    """
    Finds blobs in the given grayscale image using the Difference of Gaussians
    (DoG) method.

    Args:
        data: The grayscale input image.
        min_sigma: The minimum blob radius in pixels.
        max_sigma: The maximum blob radius in pixels.
        overlap: The minimum overlap between two blobs in pixels. For instance, if two
         blobs are detected with radii of 10 and 12 respectively, and the ``overlap``
         is set to 0.5, then the area of the smaller blob will be ignored and only the
         area of the larger blob will be returned.
        threshold_rel: The absolute lower bound for scale space maxima. Local maxima
         smaller than ``threshold_rel`` are ignored. Reduce this to detect blobs with
         less intensities.
        exclude_border: If ``True``, exclude blobs from detection if they are too
         close to the border of the image. Border size is ``min_sigma``.

    Returns:
        Coordinates of blobs
    """
    # Use scikit-image's Difference of Gaussians (DoG) method
    blobs = feature.blob_dog(
        data,
        min_sigma=min_sigma,
        max_sigma=max_sigma,
        overlap=overlap,
        threshold_rel=threshold_rel,
        exclude_border=exclude_border,
    )
    return __blobs_to_coords(blobs)


def find_blobs_doh(
    data: np.ndarray,
    min_sigma: float = 1,
    max_sigma: float = 30,
    overlap: float = 0.5,
    log_scale: bool = False,
    threshold_rel: float = 0.2,
) -> np.ndarray:
    """
    Finds blobs in the given grayscale image using the Determinant of Hessian
    (DoH) method.

    Args:
        data: The grayscale input image.
        min_sigma: The minimum blob radius in pixels.
        max_sigma: The maximum blob radius in pixels.
        overlap: The minimum overlap between two blobs in pixels. For instance, if two
         blobs are detected with radii of 10 and 12 respectively, and the ``overlap``
         is set to 0.5, then the area of the smaller blob will be ignored and only the
         area of the larger blob will be returned.
        log_scale: If ``True``, the radius of each blob is returned as ``sqrt(sigma)``
         for each detected blob.
        threshold_rel: The absolute lower bound for scale space maxima. Local maxima
         smaller than ``threshold_rel`` are ignored. Reduce this to detect blobs with
         less intensities.

    Returns:
        Coordinates of blobs
    """
    # Use scikit-image's Determinant of Hessian (DoH) method to detect blobs
    blobs = feature.blob_doh(
        data,
        min_sigma=min_sigma,
        max_sigma=max_sigma,
        num_sigma=int(max_sigma - min_sigma + 1),
        threshold=None,
        threshold_rel=threshold_rel,
        overlap=overlap,
        log_scale=log_scale,
    )
    return __blobs_to_coords(blobs)


def find_blobs_log(
    data: np.ndarray,
    min_sigma: float = 1,
    max_sigma: float = 30,
    overlap: float = 0.5,
    log_scale: bool = False,
    threshold_rel: float = 0.2,
    exclude_border: bool = True,
) -> np.ndarray:
    """Finds blobs in the given grayscale image using the Laplacian of Gaussian
    (LoG) method.

    Args:
        data: The grayscale input image.
        min_sigma: The minimum blob radius in pixels.
        max_sigma: The maximum blob radius in pixels.
        overlap: The minimum overlap between two blobs in pixels. For instance, if
         two blobs are detected with radii of 10 and 12 respectively, and the
         ``overlap`` is set to 0.5, then the area of the smaller blob will be ignored
         and only the area of the larger blob will be returned.
        log_scale: If ``True``, the radius of each blob is returned as ``sqrt(sigma)``
         for each detected blob.
        threshold_rel: The absolute lower bound for scale space maxima. Local maxima
         smaller than ``threshold_rel`` are ignored. Reduce this to detect blobs with
         less intensities.
        exclude_border: If ``True``, exclude blobs from detection if they are too
         close to the border of the image. Border size is ``min_sigma``.

    Returns:
        Coordinates of blobs
    """
    # Use scikit-image's Laplacian of Gaussian (LoG) method to detect blobs
    blobs = feature.blob_log(
        data,
        min_sigma=min_sigma,
        max_sigma=max_sigma,
        num_sigma=int(max_sigma - min_sigma + 1),
        threshold=None,
        threshold_rel=threshold_rel,
        overlap=overlap,
        log_scale=log_scale,
        exclude_border=exclude_border,
    )
    return __blobs_to_coords(blobs)


def remove_overlapping_disks(coords: np.ndarray) -> np.ndarray:
    """Remove overlapping disks among coordinates

    Args:
        coords: The coordinates of the disks

    Returns:
        The coordinates of the disks with overlapping disks removed
    """
    # Get the radii of each disk from the coordinates
    radii = coords[:, 2]
    # Calculate the distance between the center of each pair of disks
    dist = np.sqrt(np.sum((coords[:, None, :2] - coords[:, :2]) ** 2, axis=-1))
    # Create a boolean mask where the distance between the centers
    # is less than the sum of the radii
    mask = dist < (radii[:, None] + radii)
    # Find the indices of overlapping disks
    overlapping_indices = np.argwhere(mask)
    # Remove the smaller disk from each overlapping pair
    for i, j in overlapping_indices:
        if i != j:
            if radii[i] < radii[j]:
                coords[i] = [np.nan, np.nan, np.nan]
            else:
                coords[j] = [np.nan, np.nan, np.nan]
    # Remove rows with NaN values
    coords = coords[~np.isnan(coords).any(axis=1)]
    return coords


def find_blobs_opencv(
    data: np.ndarray,
    min_threshold: float | None = None,
    max_threshold: float | None = None,
    min_repeatability: int | None = None,
    min_dist_between_blobs: float | None = None,
    filter_by_color: bool | None = None,
    blob_color: int | None = None,
    filter_by_area: bool | None = None,
    min_area: float | None = None,
    max_area: float | None = None,
    filter_by_circularity: bool | None = None,
    min_circularity: float | None = None,
    max_circularity: float | None = None,
    filter_by_inertia: bool | None = None,
    min_inertia_ratio: float | None = None,
    max_inertia_ratio: float | None = None,
    filter_by_convexity: bool | None = None,
    min_convexity: float | None = None,
    max_convexity: float | None = None,
) -> np.ndarray:
    """
    Finds blobs in the given grayscale image using OpenCV's SimpleBlobDetector.

    Args:
        data: The grayscale input image.
        min_threshold: The minimum blob intensity.
        max_threshold: The maximum blob intensity.
        min_repeatability: The minimum number of times a blob is detected
         before it is reported.
        min_dist_between_blobs: The minimum distance between blobs.
        filter_by_color: If ``True``, blobs are filtered by color.
        blob_color: The color of the blobs to filter by.
        filter_by_area: If ``True``, blobs are filtered by area.
        min_area: The minimum blob area.
        max_area: The maximum blob area.
        filter_by_circularity: If ``True``, blobs are filtered by circularity.
        min_circularity: The minimum blob circularity.
        max_circularity: The maximum blob circularity.
        filter_by_inertia: If ``True``, blobs are filtered by inertia.
        min_inertia_ratio: The minimum blob inertia ratio.
        max_inertia_ratio: The maximum blob inertia ratio.
        filter_by_convexity: If ``True``, blobs are filtered by convexity.
        min_convexity: The minimum blob convexity.
        max_convexity: The maximum blob convexity.

    Returns:
        Coordinates of blobs
    """

    # Note:
    # Importing OpenCV inside the function in order to eventually raise an ImportError
    # when the function is called and OpenCV is not installed. This error will be
    # handled by DataLab and the user will be informed that OpenCV is required to use
    # this function.
    import cv2  # pylint: disable=import-outside-toplevel

    params = cv2.SimpleBlobDetector_Params()
    if min_threshold is not None:
        params.minThreshold = min_threshold
    if max_threshold is not None:
        params.maxThreshold = max_threshold
    if min_repeatability is not None:
        params.minRepeatability = min_repeatability
    if min_dist_between_blobs is not None:
        params.minDistBetweenBlobs = min_dist_between_blobs
    if filter_by_color is not None:
        params.filterByColor = filter_by_color
    if blob_color is not None:
        params.blobColor = blob_color
    if filter_by_area is not None:
        params.filterByArea = filter_by_area
    if min_area is not None:
        params.minArea = min_area
    if max_area is not None:
        params.maxArea = max_area
    if filter_by_circularity is not None:
        params.filterByCircularity = filter_by_circularity
    if min_circularity is not None:
        params.minCircularity = min_circularity
    if max_circularity is not None:
        params.maxCircularity = max_circularity
    if filter_by_inertia is not None:
        params.filterByInertia = filter_by_inertia
    if min_inertia_ratio is not None:
        params.minInertiaRatio = min_inertia_ratio
    if max_inertia_ratio is not None:
        params.maxInertiaRatio = max_inertia_ratio
    if filter_by_convexity is not None:
        params.filterByConvexity = filter_by_convexity
    if min_convexity is not None:
        params.minConvexity = min_convexity
    if max_convexity is not None:
        params.maxConvexity = max_convexity
    detector = cv2.SimpleBlobDetector_create(params)
    image = exposure.rescale_intensity(data, out_range=np.uint8)
    keypoints = detector.detect(image)
    if keypoints:
        coords = cv2.KeyPoint_convert(keypoints)
        radii = 0.5 * np.array([kp.size for kp in keypoints])
        blobs = np.vstack([coords[:, 1], coords[:, 0], radii]).T
        blobs = remove_overlapping_disks(blobs)
    else:
        blobs = np.array([]).reshape((0, 3))
    return __blobs_to_coords(blobs)
